# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import sys
import unittest
import unittest.mock as mock
from typing import Any
from typing import Dict
from unittest import TestCase

import numpy as np
import pandas as pd
from parameterized.parameterized import parameterized

from kats.consts import TimeSeriesData
from kats.data.utils import load_air_passengers
from kats.data.utils import load_data
from kats.models import arima
from kats.models import holtwinters
from kats.models import linear_model
from kats.models import quadratic_model
from kats.models import sarima
from kats.models import theta
from kats.models.ensemble.ensemble import BaseEnsemble
from kats.models.ensemble.ensemble import BaseModelParams
from kats.models.ensemble.ensemble import EnsembleParams
from kats.models.ensemble.kats_ensemble import KatsEnsemble
from kats.models.ensemble.median_ensemble import MedianEnsembleModel
from kats.models.ensemble.weighted_avg_ensemble import WeightedAvgEnsemble
from kats.models.model import Model

np.random.seed(123321)
DATA_dummy = pd.DataFrame(
    {
        "time": pd.date_range(start="2019-01-01", end="2019-12-31", freq="D"),
        "y": [x + np.random.randint(20) for x in range(365)],
    }
)

ALL_ERRORS = ["mape", "smape", "mae", "mase", "mse", "rmse"]


def get_fake_preds(ts_data: TimeSeriesData, fcst_periods: int, fcst_freq: str) -> pd.DataFrame:
    time = pd.date_range(start=ts_data.time.iloc[-1], periods=fcst_periods + 1, freq=fcst_freq)[1:]
    fcst = np.random.uniform(0, 100, len(time))
    return pd.DataFrame(
        {
            "time": {i: t for i, t in enumerate(time)},
            "fcst": {i: t for i, t in enumerate(fcst)},
            "fcst_lower": {i: t for i, t in enumerate(fcst + 10)},
            "fcst_upper": {i: t for i, t in enumerate(fcst - 10)},
        }
    )


# pyre-fixme[24]: Generic type `Model` expects 1 type parameter.
def get_predict_model(m: Model, model_name: str, steps: int, freq: str) -> np.ndarray:
    """Get model prediction based on model_name."""
    if model_name == "BaseEnsemble":
        # pyre-fixme[16]: `Model` has no attribute `_predict_all`.
        return m._predict_all(steps=steps, freq=freq)
    else:
        # pyre-fixme[7]: Expected `ndarray` but got `None`.
        return m.predict(steps=steps, freq=freq)


def get_ensemble_param(ts_param: Dict[str, bool]) -> EnsembleParams:
    """Returns EnsembleParams based on which base_models are included."""
    base_model_list = [
        BaseModelParams("arima", arima.ARIMAParams(p=1, d=1, q=1)) if ts_param["arima"] else "",
        BaseModelParams("holtwinters", holtwinters.HoltWintersParams()) if ts_param["holtwinters"] else "",
        BaseModelParams(
            "sarima",
            sarima.SARIMAParams(
                p=2,
                d=1,
                q=1,
                trend="ct",
                seasonal_order=(1, 0, 1, 12),
                enforce_invertibility=False,
                enforce_stationarity=False,
            ),
        )
        if ts_param["sarima"]
        else "",
        BaseModelParams("linear", linear_model.LinearModelParams()) if ts_param["linear"] else "",
        BaseModelParams("quadratic", quadratic_model.QuadraticModelParams()) if ts_param["quadratic"] else "",
        BaseModelParams("theta", theta.ThetaParams(m=12)) if ts_param["theta"] else "",
    ]
    return EnsembleParams(
        # pyre-fixme[6]: For 1st param expected `List[BaseModelParams]` but got
        #  `List[Union[BaseModelParams, str]]`.
        [base_model for base_model in base_model_list if base_model != ""]
    )


# test params, True or False means that base_emodel is included in this ensemble_model_param or not.
# params_base is being used for the BaseEnsemble model, and similarly the same for the KatsEnsemble
TEST_PARAM = {
    "params_base": {
        "arima": True,
        "holtwinters": True,
        "sarima": True,
        "linear": True,
        "quadratic": True,
        "theta": False,
    },
    "params_multivariate_data": {
        "arima": True,
        "holtwinters": True,
        "sarima": False,
        "linear": False,
        "quadratic": False,
        "theta": False,
    },
    "params_kats": {
        "arima": True,
        "holtwinters": False,
        "sarima": True,
        "linear": True,
        "quadratic": True,
        "theta": True,
    },
}


TEST_DATA: Dict[str, Dict[str, Any]] = {
    "monthly": {
        "ts": load_air_passengers(),
        "params": {
            "base": get_ensemble_param(TEST_PARAM["params_base"]),
            "kats": get_ensemble_param(TEST_PARAM["params_kats"]),
        },
    },
    "daily": {
        "ts": TimeSeriesData(load_data("peyton_manning.csv").set_axis(["time", "y"], axis=1)),
        "params": {
            "base": get_ensemble_param(TEST_PARAM["params_base"]),
            "kats": get_ensemble_param(TEST_PARAM["params_kats"]),
        },
    },
    "multivariate": {
        "ts": TimeSeriesData(load_data("multivariate_anomaly_simulated_data.csv")),
        "params": get_ensemble_param(TEST_PARAM["params_multivariate_data"]),
    },
    "dummy": {
        "ts": TimeSeriesData(DATA_dummy),
        "params": {
            "base": get_ensemble_param(TEST_PARAM["params_base"]),
            "kats": get_ensemble_param(TEST_PARAM["params_kats"]),
        },
    },
}


class testEnsembleModels(TestCase):
    """Test Three models BaseEnsemble, MedianEnsembleModel and WeightedAvgEnsemble."""

    def test_valid_params_base_params(self) -> None:
        """Check using a valid model params for BaseModelParams."""
        # test validate_param in base params
        base_param = BaseModelParams("arima", arima.ARIMAParams(p=1, d=1, q=1))
        base_param.validate_params()

    def test_invalid_params_in_ensemble_params(self) -> None:
        """Check using a non-valid model name in EnsembleParams results in error."""
        # validate params in EnsembleParams
        TSData = TEST_DATA["monthly"]["ts"]
        params = EnsembleParams(
            [
                BaseModelParams("random_model_name", arima.ARIMAParams(p=1, d=1, q=1)),
                BaseModelParams("holtwinters", holtwinters.HoltWintersParams()),
            ]
        )
        self.assertRaises(
            ValueError,
            BaseEnsemble,
            TSData,
            params,
        )

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            [
                "BaseEnsemble_monthly",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["params"]["base"],
                30,
                "MS",
                BaseEnsemble,
                False,
            ],
            [
                "BaseEnsemble_daily",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["params"]["base"],
                30,
                "D",
                BaseEnsemble,
                False,
            ],
            [
                "BaseEnsemble_dummy",
                TEST_DATA["dummy"]["ts"],
                TEST_DATA["dummy"]["params"]["base"],
                30,
                "D",
                BaseEnsemble,
                False,
            ],
            [
                "MedianEnsembleModel_monthly",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["params"]["base"],
                30,
                "MS",
                MedianEnsembleModel,
                False,
            ],
            [
                "MedianEnsembleModel_daily",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["params"]["base"],
                30,
                "D",
                MedianEnsembleModel,
                False,
            ],
            [
                "MedianEnsembleModel_dummy",
                TEST_DATA["dummy"]["ts"],
                TEST_DATA["dummy"]["params"]["base"],
                30,
                "D",
                MedianEnsembleModel,
                False,
            ],
            [
                "WeightedAvgEnsemble_monthly",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["params"]["base"],
                30,
                "MS",
                WeightedAvgEnsemble,
                False,
            ],
            [
                "WeightedAvgEnsemble_daily",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["params"]["base"],
                30,
                "D",
                WeightedAvgEnsemble,
                False,
            ],
            [
                "WeightedAvgEnsemble_dummy",
                TEST_DATA["dummy"]["ts"],
                TEST_DATA["dummy"]["params"]["base"],
                30,
                "D",
                WeightedAvgEnsemble,
                True,
            ],
        ]
    )
    def test_forecast(
        self,
        ts_data_name: str,
        ts_data: TimeSeriesData,
        # pyre-fixme[2]: Parameter must be annotated.
        params,
        steps: int,
        # pyre-fixme[2]: Parameter must be annotated.
        freq,
        # pyre-fixme[2]: Parameter must be annotated.
        model,
        backtester: bool,
    ) -> None:
        """Test forecast."""
        preds = get_fake_preds(ts_data, fcst_periods=steps, fcst_freq=freq)[["time", "fcst"]]

        m = model(data=ts_data, params=params)

        with mock.patch("kats.models.ensemble.ensemble.Pool") as mock_pooled:
            mock_fit_model = mock_pooled.return_value.apply_async.return_value.get
            mock_fit_model.return_value.predict = mock.MagicMock(return_value=preds)

            # fit the ensemble model
            m.fit()
            mock_pooled.assert_called()

            if backtester:
                with mock.patch("kats.models.ensemble.weighted_avg_ensemble.Pool") as mock_weighted_pooled:
                    mock_backtest = mock_weighted_pooled.return_value.apply_async.return_value.get
                    # the backtester should just return a random number here
                    mock_backtest.return_value = np.random.rand()
                    m.predict(steps=steps, freq=freq)
                    mock_backtest.assert_called()
            else:
                mock_fit_model.assert_called()
                # no predictions should be made yet
                mock_fit_model.return_value.predict.assert_not_called()

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            [
                "BaseEnsemble",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["params"]["base"],
                30,
                "MS",
                BaseEnsemble,
                False,
                False,
            ],
            [
                "MedianEnsembleModel",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["params"]["base"],
                30,
                "MS",
                MedianEnsembleModel,
                False,
                True,
            ],
            [
                "WeightedAvgEnsemble",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["params"]["base"],
                30,
                "MS",
                WeightedAvgEnsemble,
                True,
                True,
            ],
        ]
    )
    def test_predict_plot(
        self,
        ts_model_name: str,
        ts_data: TimeSeriesData,
        # pyre-fixme[2]: Parameter must be annotated.
        params,
        steps: int,
        # pyre-fixme[2]: Parameter must be annotated.
        freq,
        # pyre-fixme[2]: Parameter must be annotated.
        model,
        backtester: bool,
        plot: bool,
    ) -> None:
        m = model(data=ts_data, params=params)
        preds = get_fake_preds(ts_data, fcst_periods=steps, fcst_freq=freq)[["time", "fcst"]]
        with mock.patch("kats.models.ensemble.ensemble.Pool") as mock_pooled:
            mock_fit_model = mock_pooled.return_value.apply_async.return_value.get
            mock_fit_model.return_value.predict = mock.MagicMock(return_value=preds)

            # fit the ensemble model
            m.fit()
            mock_pooled.assert_called()
            if backtester:
                with mock.patch("kats.models.ensemble.weighted_avg_ensemble.Pool") as mock_weighted_pooled:
                    mock_backtest = mock_weighted_pooled.return_value.apply_async.return_value.get
                    # the backtester should just return a random number here
                    mock_backtest.return_value = np.random.rand()
                    m.predict(steps=steps, freq=freq)
                    mock_backtest.assert_called()
                    if plot:
                        m.plot()
            else:
                mock_fit_model.assert_called()
                # no predictions should be made yet
                mock_fit_model.return_value.predict.assert_not_called()
                # now run predict on the ensemble model
                get_predict_model(m, ts_model_name, steps=steps, freq=freq)
                mock_fit_model.return_value.predict.assert_called_with(steps, freq=f"{freq}")
                if plot:
                    m.plot()

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            [
                "BaseEnsemble",
                BaseEnsemble,
                "Ensemble",
            ],
            [
                "MedianEnsembleModel",
                MedianEnsembleModel,
                "Median Ensemble",
            ],
            [
                "WeightedAvgEnsemble",
                WeightedAvgEnsemble,
                "Weighted Average Ensemble",
            ],
        ]
    )
    # pyre-fixme[2]: Parameter must be annotated.
    def test_name(self, ts_model_name: str, model, model_name: str) -> None:
        """Test name of the model according to the model used."""
        # test __str__ method
        m = model(TEST_DATA["daily"]["ts"], TEST_DATA["daily"]["params"]["base"])
        self.assertEqual(m.__str__(), model_name)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            [
                "BaseEnsemble",
                BaseEnsemble,
            ],
            [
                "MedianEnsembleModel",
                MedianEnsembleModel,
            ],
            [
                "WeightedAvgEnsemble",
                WeightedAvgEnsemble,
            ],
        ]
    )
    def test_invalid_params_ensemble_params(
        self,
        ts_model_name: str,
        # pyre-fixme[24]: Generic type `Model` expects 1 type parameter.
        model: Model,
    ) -> None:
        # validate params in EnsembleParams
        TSData_multi = TEST_DATA["multivariate"]["ts"]
        params = TEST_DATA["multivariate"]["params"]

        self.assertRaises(
            ValueError,
            # pyre-fixme[6]: For 2nd param expected `(...) -> Any` but got
            #  `Model[typing.Any]`.
            model,
            TSData_multi,
            params,
        )


class testKatsEnsemble(TestCase):
    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            [
                "monthly_kats",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["params"]["kats"],
                30,
                "MS",
            ],
            [
                "daily_kats",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["params"]["kats"],
                30,
                "D",
            ],
            [
                "dummy_kats",
                TEST_DATA["dummy"]["ts"],
                TEST_DATA["dummy"]["params"]["kats"],
                30,
                "D",
            ],
        ]
    )
    def test_fit_median_forecast(
        self,
        ts_data_name: str,
        ts_data: TimeSeriesData,
        params: Dict[str, Any],
        steps: int,
        freq: str,
    ) -> None:
        preds = get_fake_preds(ts_data, fcst_periods=steps, fcst_freq=freq)

        decomps = ["additive", "multiplicative"]

        for decomp in decomps:
            KatsEnsembleParam = {
                "models": params,
                "aggregation": "median",
                "seasonality_length": 12,
                "decomposition_method": decomp,
            }

            m = KatsEnsemble(data=ts_data, params=KatsEnsembleParam)

            with mock.patch("multiprocessing.managers.SyncManager.Pool") as mock_pooled:
                mock_fit_model = mock_pooled.return_value.apply_async.return_value.get
                mock_fit_model.return_value.predict = mock.MagicMock(return_value=preds)
                # fit the model
                m.fit()
                mock_pooled.assert_called()
                # no predictions should be made yet
                mock_fit_model.return_value.predict.assert_not_called()
                # now run predict on the ensemble model
                m.predict(steps=steps)
                mock_fit_model.return_value.predict.assert_called_with(steps)
                m.aggregate()
                m.plot()

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            [
                "monthly",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["params"]["kats"],
                30,
                "MS",
            ],
            [
                "daily",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["params"]["kats"],
                30,
                "D",
            ],
            [
                "dummy",
                TEST_DATA["dummy"]["ts"],
                TEST_DATA["dummy"]["params"]["kats"],
                30,
                "D",
            ],
        ]
    )
    def test_fit_weightedavg_forecast(
        self,
        ts_data_name: str,
        ts_data: TimeSeriesData,
        params: Dict[str, Any],
        steps: int,
        freq: str,
    ) -> None:
        preds = get_fake_preds(ts_data, fcst_periods=steps, fcst_freq=freq)

        decomps = ["additive", "multiplicative"]

        for decomp in decomps:
            KatsEnsembleParam = {
                "models": params,
                "aggregation": "weightedavg",
                "seasonality_length": 12,
                "decomposition_method": decomp,
            }
            m = KatsEnsemble(data=ts_data, params=KatsEnsembleParam)

            with mock.patch("multiprocessing.managers.SyncManager.Pool") as mock_pooled:
                mock_fit_model = mock_pooled.return_value.apply_async.return_value.get
                mock_fit_model.return_value.predict = mock.MagicMock(return_value=preds)
                mock_fit_model.return_value.__add__ = mock.MagicMock(return_value=np.random.rand())
                # fit the model
                m.fit()
                mock_pooled.assert_called()

                # no predictions should be made yet
                mock_fit_model.return_value.predict.assert_not_called()
                # backtesting should be done after calling fit
                mock_fit_model.return_value.__add__.assert_called_with(sys.float_info.epsilon)

                # now run predict on the ensemble model
                m.predict(steps=steps)
                mock_fit_model.return_value.predict.assert_called_with(steps)
                m.aggregate()
                m.plot()

                # reset all the mocks and make sure they're not called
                mock_pooled.reset_mock()
                mock_pooled.assert_not_called()
                mock_fit_model.return_value.predict.assert_not_called()
                mock_fit_model.return_value.__add__.assert_not_called()

                # now retry the above with forecast rather than fit/predict
                m.forecast(steps=30)
                mock_pooled.assert_called()

                # backtesting should be done after calling fit
                mock_fit_model.return_value.__add__.assert_called_with(sys.float_info.epsilon)

                # now run predict on the ensemble model
                m.predict(steps=steps)
                mock_fit_model.return_value.predict.assert_called_with(steps)
                m.aggregate()
                m.plot()

    def test_invalid_decomposition_method(self) -> None:
        KatsEnsembleParam = {
            "models": get_ensemble_param(TEST_PARAM["params_kats"]),
            "aggregation": "median",
            "seasonality_length": 12,
            "decomposition_method": "random_decomp",
        }
        m = KatsEnsemble(data=TEST_DATA["monthly"]["ts"], params=KatsEnsembleParam)
        m.validate_params()

    def test_invalid_seasonality_length(self) -> None:
        KatsEnsembleParam = {
            "models": get_ensemble_param(TEST_PARAM["params_kats"]),
            "aggregation": "median",
            "seasonality_length": 1000000,
            "decomposition_method": "additive",
        }

        self.assertRaises(
            ValueError,
            KatsEnsemble,
            TEST_DATA["monthly"]["ts"],
            KatsEnsembleParam,
        )

    def test_logging_default_executors(self) -> None:
        KatsEnsembleParam = {
            "models": get_ensemble_param(TEST_PARAM["params_kats"]),
            "aggregation": "median",
            "seasonality_length": 12,
            "decomposition_method": "random_decomp",
            "fitExecutor": None,
            "forecastExecutor": None,
        }
        with self.assertLogs(level="INFO"):
            KatsEnsemble(data=TEST_DATA["monthly"]["ts"], params=KatsEnsembleParam)

    def test_non_seasonal_data(self) -> None:
        KatsEnsembleParam = {
            "models": get_ensemble_param(TEST_PARAM["params_kats"]),
            "aggregation": "median",
            "seasonality_length": 12,
            "decomposition_method": "additive",
        }

        dummy_ts = TimeSeriesData(
            time=pd.date_range(start="2020-01-01", end="2020-05-31", freq="D"),
            value=pd.Series(list(range(152))),
        )
        KatsEnsemble(data=dummy_ts, params=KatsEnsembleParam)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
from datetime import datetime
from typing import Any
from typing import Optional
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.stats as stats

from kats.consts import IRREGULAR_GRANULARITY_ERROR
from kats.consts import TimeSeriesData
from kats.detectors.detector import DetectorModel
from kats.detectors.detector_consts import AnomalyResponse
from kats.detectors.detector_consts import ChangePointInterval
from kats.detectors.detector_consts import ConfidenceBand
from kats.detectors.detector_consts import PercentageChange
from kats.utils.decomposition import TimeSeriesDecomposition

"""Statistical Significance Detector Module
This module contains simple detectors that apply a t-test over a rolling window to compare
check if there is a statistically significant increase or decrease between the control and test
time periods.  In addition to the univariate version of this test, this module includes a
multivariate version that uses a false discovery rate (FDR) controlling procedure to reduce noise.
"""


class StatSigDetectorModel(DetectorModel):
    """
    StatSigDetectorModel is a simple detector, which compares a control and test period.
    The detector assumes that the time series data comes from a iid normal distribution,
    and applies a t-test to check if the means between the control and test period are
    significantly different.

    We start with the history data, and then as for the current data, we apply a rolling
    window, adding one data point at a time from the current data, and detecting significant
    change. We return the t-statistic as a score, which reflects the severity of the
    change.
    We suggest using n_control >= 30 to get good estimates

    Attributes:
        n_control: number of data points(or time units) of history to compare with
        n_test: number of points(or time_units) to compare the history with
        serialized_model: serialized json containing the parameters
        time_units: units of time used to measure the intervals. If not provided
                    we infer it from the provided data.
        rem_season: default value is False, if remove seasonality for historical data and data
        seasonal_period: str, default value is 'weekly'. Other possible values: 'daily', 'biweekly', 'monthly', 'yearly'
        use_corrected_scores: bool, default value is False, using original t-scores or correct t-scores.
        max_split_ts_length: int, default value is 500. If the given TS (except historical part) is longer than max_split_ts_length,
                    we will transform a long univariate TS into a multi-variate TS and then use multistatsig detector, which is faster,
        anomaly_scores_only: bool = False. Only calculate anomaly scores without using advanced classes, which is much faster.
        min_perc_change: float, minimum percentage change, for a non zero score. Score will be clipped to zero if the absolute value of the percentage chenge is less than this value
    >>> # Example usage:
    >>> # history and ts_pt are TimeSeriesData objects and history is larger
    >>> # than (n_control + n_test) so that we have sufficient history to
    >>> # run the detector
    >>> n_control = 28
    >>> n_test = 7
    >>> import random
    >>> control_time = pd.date_range(start='2018-01-01', freq='D', periods=(n_control + n_test))
    >>> test_time = pd.date_range(start='2018-02-05', freq='D', periods=n_test)
    >>> control_val = [random.normalvariate(100,10) for _ in range(n_control + n_test)]
    >>> test_val = [random.normalvariate(120,10) for _ in range(n_test)]
    >>> hist_ts = TimeSeriesData(time=control_time, value=pd.Series(control_val))
    >>> data_ts = TimeSeriesData(time=test_time, value=pd.Series(test_val))
    >>> ss_detect = StatSigDetectorModel(n_control=n_control, n_test=n_test)
    >>> anom = ss_detect.fit_predict(data=data_ts, historical_data=hist_ts)
    >>> # if only caculate anomaly scores
    >>> ss_detect = StatSigDetectorModel(n_control=n_control, n_test=n_test, anomaly_scores_only=True)
    >>> anom = ss_detect.fit_predict(data=data_ts, historical_data=hist_ts)
    """

    data: Optional[TimeSeriesData] = None

    def __init__(
        self,
        n_control: Optional[int] = None,
        n_test: Optional[int] = None,
        serialized_model: Optional[bytes] = None,
        time_unit: Optional[str] = None,
        rem_season: bool = False,
        seasonal_period: str = "weekly",
        use_corrected_scores: bool = True,
        max_split_ts_length: int = 500,
        anomaly_scores_only: bool = False,
        min_perc_change: float = 0.0,
    ) -> None:

        if serialized_model:
            model_dict = json.loads(serialized_model)
            self.n_test: int = model_dict["n_test"]
            self.n_control: int = model_dict["n_control"]
            self.time_unit: str = model_dict["time_unit"]
            # for seasonality
            self.rem_season: bool = model_dict.get("rem_season", rem_season)
            self.seasonal_period: str = model_dict.get("seasonal_period", seasonal_period)

            # for big data and correct t-scores
            self.use_corrected_scores: bool = model_dict.get("use_corrected_scores", use_corrected_scores)
            # threshold for splitting long TS
            self.max_split_ts_length: int = model_dict.get("max_split_ts_length", max_split_ts_length)

            self.min_perc_change: float = model_dict.get("min_perc_change", min_perc_change)

        else:
            self.n_test: Optional[int] = n_test
            self.n_control: Optional[int] = n_control
            self.time_unit: Optional[str] = time_unit
            # for seasonality
            self.rem_season: bool = rem_season
            self.seasonal_period: str = seasonal_period
            # big data and t-scores
            self.use_corrected_scores: bool = use_corrected_scores
            # threshold for splitting long TS
            self.max_split_ts_length: int = max_split_ts_length
            self.min_perc_change: float = min_perc_change

        if (self.n_control is None) or (self.n_test is None):
            raise ValueError("You must either provide serialized model or values for control " "and test intervals.")

        self.control_interval: Optional[ChangePointInterval] = None
        self.test_interval: Optional[ChangePointInterval] = None
        self.response: Optional[AnomalyResponse] = None

        self.is_initialized = False  # flag on whether initialized or not
        self.last_N = 0  # this is the size of the last chunk of data we saw
        self.data_history: Optional[TimeSeriesData] = None

        # for seasonality
        self.data_season: Optional[TimeSeriesData] = None

        # big data strategy
        self.bigdata_trans_flag: Optional[bool] = None
        self.remaining: Optional[int] = None

        # validate time_unit
        if self.time_unit is not None:
            try:
                # for digit+str ('3D', '10h') cases
                _ = pd.Timedelta(self.time_unit)
            except ValueError:
                try:
                    _ = pd.Timedelta(1, unit=self.time_unit)
                    assert self.time_unit is not None
                    self.time_unit = "1" + self.time_unit
                    _ = pd.Timedelta(self.time_unit)
                except ValueError:
                    raise ValueError("Invalide time_unit value.")

        self.anomaly_scores_only: bool = anomaly_scores_only

    def serialize(self) -> bytes:
        """
        Serializes by putting model parameters in a json
        """
        model_dict = {
            "n_control": self.n_control,
            "n_test": self.n_test,
            "time_unit": self.time_unit,
            "rem_season": self.rem_season,
            "seasonal_period": self.seasonal_period,
            "use_corrected_scores": self.use_corrected_scores,
            "max_split_ts_length": self.max_split_ts_length,
            "min_perc_change": self.min_perc_change,
        }

        return json.dumps(model_dict).encode("utf-8")

    def fit_predict(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData] = None,
        **kwargs: Any,
    ) -> AnomalyResponse:
        """
        This is the main working function.
        The function returns an AnomalyResponse object of length
        equal to the length of the data.
        We require len(historical_data) > (n_control + n_test).

        Args:
            data: TimeSeriesData, A univariate TimeSeriesData for which we are running the StatSigDetectorModel
            historical_data: Optional[TimeSeriesData] Historical data used to do detection for initial points in data

        """

        if not data.is_univariate():
            msg = "Input is multivariate but StatSigDetector expected univariate input."
            logging.error(msg)
            raise ValueError(msg)

        self._set_time_unit(data=data, historical_data=historical_data)
        self.last_N = len(data)

        # this ensures we start with a default response of
        # the size of the data
        self._init_response(data)
        response = self.response
        assert response is not None

        # when there is no need to update
        # just return the initial response of zeros
        if not self._should_update(data=data, historical_data=historical_data):
            return response

        # if we only need anomaly scores
        if self.anomaly_scores_only:
            return self._get_anomaly_scores_only(
                historical_data=historical_data,
                data=data,
            )

        # handle cases where there is either no historical data, or
        # not enough historical data
        data, historical_data = self._handle_not_enough_history(
            data=data,
            historical_data=historical_data,
        )

        # remove seasonality
        if self.rem_season:
            sh_data = SeasonalityHandler(data=data, seasonal_period=self.seasonal_period)

            self.data_season = sh_data.get_seasonality()
            data = sh_data.remove_seasonality()

            if historical_data:
                sh_hist_data = SeasonalityHandler(
                    data=historical_data,
                    seasonal_period=self.seasonal_period,
                )
                historical_data = sh_hist_data.remove_seasonality()

        self.data = data
        # first initialize this with the historical data
        self._init_data(historical_data)

        # if using new t-scores
        if self.use_corrected_scores:
            if (
                len(data) > self.max_split_ts_length
                # pyre-ignore[16]: `Optional` has no attribute `infer_freq_robust`.
                and historical_data.infer_freq_robust() == data.infer_freq_robust()
            ):
                self.bigdata_trans_flag = True
            else:
                self.bigdata_trans_flag = False
        else:
            self.bigdata_trans_flag = False

        # if need trans to multi-TS
        if self.bigdata_trans_flag:
            new_data_ts = self._reorganize_big_data(self.max_split_ts_length)
            ss_detect = MultiStatSigDetectorModel(
                n_control=self.n_control,
                n_test=self.n_test,
                time_unit=self.time_unit,
                rem_season=False,
                seasonal_period=self.seasonal_period,
                skip_rescaling=True,
                use_corrected_scores=self.use_corrected_scores,
            )
            anom = ss_detect.fit_predict(data=new_data_ts)
            self._reorganize_back(anom)

        else:
            self._init_control_test(data if historical_data is None else historical_data)
            # set the flag to true
            self.is_initialized = True

            # now run through the data to get the prediction
            for i in range(len(data)):
                current_time = data.time.iloc[i]

                ts_pt = TimeSeriesData(
                    time=pd.Series(current_time, copy=False),
                    value=pd.Series(data.value.iloc[i], copy=False),
                )

                self._update_data(ts_pt)
                self._update_control_test(ts_pt)
                self._update_response(ts_pt.time.iloc[0])

        # add seasonality back
        if self.rem_season:
            data_season = self.data_season
            confidence_band = response.confidence_band
            predicted_ts = response.predicted_ts
            assert data_season is not None
            assert confidence_band is not None
            assert predicted_ts is not None

            start_idx = len(response.scores) - len(data_season)
            datatime = response.scores.time

            self.response = AnomalyResponse(
                scores=response.scores,
                confidence_band=ConfidenceBand(
                    upper=TimeSeriesData(
                        time=datatime,
                        value=pd.concat(
                            [
                                pd.Series(
                                    confidence_band.upper.value.values[:start_idx],
                                    copy=False,
                                ),
                                pd.Series(
                                    np.asarray(confidence_band.upper.value.values[start_idx:])
                                    + np.asarray(data_season.value.values),
                                    copy=False,
                                ),
                            ],
                            copy=False,
                        ),
                    ),
                    lower=TimeSeriesData(
                        time=datatime,
                        value=pd.concat(
                            [
                                pd.Series(
                                    confidence_band.lower.value.values[:start_idx],
                                    copy=False,
                                ),
                                pd.Series(
                                    np.asarray(confidence_band.lower.value.values[start_idx:])
                                    + np.asarray(data_season.value.values),
                                    copy=False,
                                ),
                            ],
                            copy=False,
                        ),
                    ),
                ),
                predicted_ts=TimeSeriesData(
                    time=datatime,
                    value=pd.concat(
                        [
                            pd.Series(predicted_ts.value.values[:start_idx], copy=False),
                            pd.Series(
                                np.asarray(predicted_ts.value.values[start_idx:])
                                + np.asarray(data_season.value.values),
                                copy=False,
                            ),
                        ],
                        copy=False,
                    ),
                ),
                anomaly_magnitude_ts=response.anomaly_magnitude_ts,
                stat_sig_ts=response.stat_sig_ts,
            )
        assert self.response is not None
        return self.response.get_last_n(self.last_N)

    def _get_anomaly_scores_only(
        self,
        historical_data: Optional[TimeSeriesData],
        data: TimeSeriesData,
    ) -> AnomalyResponse:
        if historical_data is None:
            total_data = data
        else:
            historical_data.extend(data, validate=False)
            total_data = historical_data

        total_data_df = total_data.to_dataframe()
        total_data_df.columns = ["time", "value"]
        total_data_df = total_data_df.set_index("time")

        res = []
        i = 0
        while i < len(data.time):
            test_end_dt = data.time[i]

            assert self.n_test is not None
            test_start_dt = test_end_dt - (self.n_test - 1) * pd.Timedelta(self.time_unit)

            control_end_dt = test_start_dt
            assert self.n_control is not None
            control_start_dt = control_end_dt - self.n_control * pd.Timedelta(self.time_unit)

            if control_end_dt in total_data_df.index:
                # exclude index control_end_dt
                group1 = total_data_df[control_start_dt:control_end_dt].value.to_list()[:-1]
            else:
                group1 = total_data_df[control_start_dt:control_end_dt].value.to_list()
            group2 = total_data_df[test_start_dt:test_end_dt].value.to_list()

            # 2 sample t-test
            if control_start_dt >= total_data.time[0]:
                res.append(stats.ttest_ind(a=group2, b=group1, equal_var=True)[0])
            else:
                res.append(0)

            i += 1

        scores = TimeSeriesData(pd.DataFrame({"time": list(data.time), "value": res}))

        return AnomalyResponse(
            scores=scores,
            confidence_band=None,
            predicted_ts=None,
            anomaly_magnitude_ts=TimeSeriesData(),
            stat_sig_ts=None,
        )

    def _reorganize_big_data(self, max_split_ts_length: int) -> TimeSeriesData:
        data_history = self.data_history
        data = self.data
        assert data_history is not None
        assert data is not None

        first_half_len = len(data_history)
        n_seq = len(data) // max_split_ts_length + int(len(data) % max_split_ts_length > 0)
        remaining = (max_split_ts_length * n_seq - len(data)) % max_split_ts_length

        time_need = pd.concat(
            [data_history.time[:], data.time[:max_split_ts_length]],
            copy=False,
        )

        new_ts = [
            list(
                pd.concat(
                    [data_history.value[:], data.value[:max_split_ts_length]],
                    copy=False,
                )
            )
        ]
        for i in range(max_split_ts_length, len(data), max_split_ts_length):
            new_ts.append(new_ts[-1][-first_half_len:] + list(data.value[i : i + max_split_ts_length]))
        new_ts[-1] += [1] * remaining

        new_data_ts = TimeSeriesData(
            pd.DataFrame(
                {
                    **{"time": time_need},
                    **{f"ts_{i}": new_ts[i] for i in range(len(new_ts))},
                },
                copy=False,
            )
        )
        self.remaining = remaining
        return new_data_ts

    def _reorganize_back(self, anom: AnomalyResponse) -> None:
        data_history = self.data_history
        remaining = self.remaining
        anom_predicted_ts = anom.predicted_ts
        anom_confidence_band = anom.confidence_band
        anom_stat_sig_ts = anom.stat_sig_ts
        response = self.response
        assert data_history is not None
        assert remaining is not None
        assert anom_predicted_ts is not None
        assert anom_confidence_band is not None
        assert anom_stat_sig_ts is not None
        assert response is not None
        response_predicted_ts = response.predicted_ts
        assert response_predicted_ts is not None

        start_point = len(data_history)
        res_score_val = pd.Series(
            pd.DataFrame(anom.scores.value, copy=False).iloc[start_point:, :].values.T.flatten()[:-remaining],
            copy=False,
        )

        res_predicted_ts_val = pd.Series(
            pd.DataFrame(anom_predicted_ts.value, copy=False).iloc[start_point:, :].values.T.flatten()[:-remaining],
            copy=False,
        )

        res_anomaly_magnitude_ts_val = pd.Series(
            pd.DataFrame(anom.anomaly_magnitude_ts.value, copy=False)
            .iloc[start_point:, :]
            .values.T.flatten()[:-remaining],
            copy=False,
        )

        res_stat_sig_ts_val = pd.Series(
            pd.DataFrame(anom_stat_sig_ts.value, copy=False).iloc[start_point:, :].values.T.flatten()[:-remaining],
            copy=False,
        )

        res_confidence_band_lower_val = pd.Series(
            pd.DataFrame(anom_confidence_band.lower.value, copy=False)
            .iloc[start_point:, :]
            .values.T.flatten()[:-remaining],
            copy=False,
        )

        res_confidence_band_upper_val = pd.Series(
            pd.DataFrame(anom_confidence_band.upper.value, copy=False)
            .iloc[start_point:, :]
            .values.T.flatten()[:-remaining],
            copy=False,
        )

        datatime = response.scores.time
        zeros = pd.Series(np.zeros(len(datatime) - len(res_score_val)), copy=False)
        datavalues = pd.Series(
            response_predicted_ts.value.values[: len(datatime) - len(res_score_val)],
            copy=False,
        )

        self.response = AnomalyResponse(
            scores=TimeSeriesData(time=datatime, value=pd.concat([zeros, res_score_val], copy=False)),
            confidence_band=ConfidenceBand(
                upper=TimeSeriesData(
                    time=datatime,
                    value=pd.concat([datavalues, res_confidence_band_upper_val], copy=False),
                ),
                lower=TimeSeriesData(
                    time=datatime,
                    value=pd.concat([datavalues, res_confidence_band_lower_val], copy=False),
                ),
            ),
            predicted_ts=TimeSeriesData(
                time=datatime,
                value=pd.concat([datavalues, res_predicted_ts_val], copy=False),
            ),
            anomaly_magnitude_ts=TimeSeriesData(
                time=datatime,
                value=pd.concat([zeros, res_anomaly_magnitude_ts_val], copy=False),
            ),
            stat_sig_ts=TimeSeriesData(time=datatime, value=pd.concat([zeros, res_stat_sig_ts_val], copy=False)),
        )

    def visualize(self) -> None:
        """Function to visualize the result of the StatSigDetectorModel."""
        self.plot()

    def plot(
        self,
        figsize: Optional[Tuple[int, int]] = None,
        line_color: str = "k",
        line_style: str = "-",
        prediction_color: str = "r",
        prediction_style: str = "--",
        fill_color: Optional[str] = "blue",
        response_color: str = "r",
        response_style: str = "-",
        **kwargs: Any,
    ) -> Tuple[plt.Axes, plt.Axes]:
        """Plot the results.

        Args:
            figsize: figsize to use. If None, use (10, 8).
            line_color: color to use for the original data line.
            line_style: linestyle to use for the original data.
            prediction_color: color to use for the predicted data.
            prediction_style: linestyle to use for the predicted data.
            fill_color: color to use for the confidence interval. If None, don't
                plot the confidence interval.
            response_color: color to use for the response plot.
            response_style: linestyle to use for the response plot.
        """
        data = self.data
        if data is None:
            raise ValueError("Call fit_predict() before visualize()")
        response = self.response
        assert response is not None
        predicted_ts = response.predicted_ts
        confidence_band = response.confidence_band
        assert predicted_ts is not None
        assert confidence_band is not None
        if figsize is None:
            figsize = (10, 8)
        _, (ax1, ax2) = plt.subplots(2, 1, figsize=figsize)
        N_data = len(data)

        ax1.plot(data.time, data.value.values, color=line_color, ls=line_style)
        ax1.plot(
            predicted_ts.time.iloc[-N_data:],
            predicted_ts.value.values[-N_data:],
            color=prediction_color,
            ls=prediction_style,
        )
        upper_ci = confidence_band.upper
        lower_ci = confidence_band.lower

        if fill_color is not None:
            ax1.fill_between(
                upper_ci.time.iloc[-N_data:],
                lower_ci.value.values[-N_data:],
                upper_ci.value.values[-N_data:],
                facecolor=fill_color,
                alpha=0.25,
            )

        ax2.plot(
            response.scores.time.iloc[-N_data:],
            response.scores.value.values[-N_data:],
            color=response_color,
            ls=response_style,
        )
        return ax1, ax2

    def _set_time_unit(self, data: TimeSeriesData, historical_data: Optional[TimeSeriesData]) -> None:
        """
        if the time unit is not set, this function tries to set it.
        """
        if not self.time_unit:
            if len(data) >= 3:
                # frequency is pd.TimeDelta
                frequency = data.infer_freq_robust()
            elif historical_data and len(historical_data) >= 3:
                frequency = historical_data.infer_freq_robust()
            else:
                raise ValueError(
                    "Not able to infer freqency of the time series. "
                    "Please use longer time series data or pass the time_unit parameter to the initializer."
                )

            # Timedelta string
            self.time_unit = f"{frequency.total_seconds()}S"

        # validate the time_unit
        try:
            # for digit+str ('3D', '10h') cases
            _ = pd.Timedelta(self.time_unit)
        except ValueError:
            try:
                _ = pd.Timedelta(1, unit=self.time_unit)
                assert self.time_unit is not None
                self.time_unit = "1" + self.time_unit
                _ = pd.Timedelta(self.time_unit)
            except ValueError:
                raise ValueError("Invalide time_unit value.")

    def _should_update(self, data: TimeSeriesData, historical_data: Optional[TimeSeriesData] = None) -> bool:
        """
        Flag on whether we should update responses, or send a default response of zeros.
        """

        if self.time_unit is None:
            raise ValueError("time_unit variable cannot be None")

        if not historical_data:
            start_time = data.time.iloc[0]
        else:
            start_time = historical_data.time.iloc[0]

        end_time = data.time.iloc[-1]
        assert self.n_control is not None
        assert self.n_test is not None

        return end_time >= (start_time + (self.n_control + self.n_test - 1) * pd.Timedelta(self.time_unit))

    def _handle_not_enough_history(
        self, data: TimeSeriesData, historical_data: Optional[TimeSeriesData]
    ) -> Tuple[TimeSeriesData, Optional[TimeSeriesData]]:
        """
        Handles the case when we don't have enough historical data.
        If we don't need to update, this does not do anything.
        If we need to update, this divides up the data accordingly.
        """
        assert self.time_unit is not None
        assert self.n_control is not None
        assert self.n_test is not None
        num_hist_points = self.n_control + self.n_test - 1

        # if we have enough history, we should not do anything
        if historical_data:
            history_first = historical_data.time.iloc[0]
            history_last = historical_data.time.iloc[-1]
            min_history_last = history_first + num_hist_points * pd.Timedelta(self.time_unit)

            if history_last >= min_history_last:
                return data, historical_data

        # when no historical data, divide the data into historical and not
        if historical_data is None:
            total_data = data
        else:
            historical_data.extend(data, validate=False)
            total_data = historical_data

        first_dt = total_data.time.iloc[0]  # first date of the data

        last_dt = first_dt + num_hist_points * pd.Timedelta(self.time_unit)

        historical_data = TimeSeriesData(
            time=total_data.time[total_data.time < last_dt],
            value=total_data.value[total_data.time < last_dt],
        )

        data = TimeSeriesData(
            time=total_data.time[total_data.time >= last_dt],
            value=total_data.value[total_data.time >= last_dt],
        )

        return data, historical_data

    def fit(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData] = None,
        **kwargs: Any,
    ) -> None:
        """
        Fit can be called during priming. It's a noop for us.
        """

        return

    def _init_response(self, data: TimeSeriesData) -> None:
        """
        Initializes a default response.
        """

        zeros = np.zeros(len(data))
        self.response = AnomalyResponse(
            scores=TimeSeriesData(time=data.time, value=pd.Series(zeros, copy=False)),
            confidence_band=ConfidenceBand(
                upper=TimeSeriesData(time=data.time, value=pd.Series(data.value.values, copy=False)),
                lower=TimeSeriesData(time=data.time, value=pd.Series(data.value.values, copy=False)),
            ),
            predicted_ts=TimeSeriesData(time=data.time, value=pd.Series(data.value.values, copy=False)),
            anomaly_magnitude_ts=TimeSeriesData(time=data.time, value=pd.Series(zeros, copy=False)),
            stat_sig_ts=TimeSeriesData(time=data.time, value=pd.Series(zeros, copy=False)),
        )

    def _update_response(self, date: pd.Timestamp) -> None:
        """
        Updates the current response with data from date.
        """
        assert self.test_interval is not None
        assert self.control_interval is not None
        perc_change = PercentageChange(
            current=self.test_interval,
            previous=self.control_interval,
            use_corrected_scores=self.use_corrected_scores,
            min_perc_change=self.min_perc_change,
        )
        assert self.response is not None
        self.response.inplace_update(
            time=date,
            score=perc_change.score,
            ci_upper=perc_change.ci_upper,
            ci_lower=perc_change.ci_lower,
            pred=perc_change.mean_previous,
            anom_mag=perc_change.mean_difference,
            stat_sig=1.0 if perc_change.stat_sig else 0.0,
        )

    def _get_start_end_dates(self, data: TimeSeriesData) -> Tuple[datetime, datetime, datetime, datetime]:
        """
        Gets the start and end dates of the initial interval.
        """

        last_dt = data.time.iloc[-1]

        test_end_dt = last_dt + pd.Timedelta(self.time_unit)

        assert self.n_test is not None
        test_start_dt = test_end_dt - self.n_test * pd.Timedelta(self.time_unit)

        assert self.n_test is not None
        assert self.n_control is not None
        control_start_dt = test_end_dt - (self.n_test + self.n_control) * pd.Timedelta(self.time_unit)

        control_end_dt = test_start_dt

        return control_start_dt, control_end_dt, test_start_dt, test_end_dt

    def _init_control_test(self, data: TimeSeriesData) -> None:
        """
        initializes the control and test intervals.
        """

        (
            control_start_dt,
            control_end_dt,
            test_start_dt,
            test_end_dt,
        ) = self._get_start_end_dates(data)
        data_history = self.data_history
        assert data_history is not None

        self.test_interval = ChangePointInterval(test_start_dt, test_end_dt)
        self.test_interval.data = data_history
        self.control_interval = ChangePointInterval(control_start_dt, control_end_dt)
        self.control_interval.data = data_history

    def _update_control_test(self, data: TimeSeriesData) -> None:
        """
        Updates control and test with new data.
        """

        (
            control_start_dt,
            control_end_dt,
            test_start_dt,
            test_end_dt,
        ) = self._get_start_end_dates(data)
        data_history = self.data_history
        assert data_history is not None
        self.test_interval = ChangePointInterval(test_start_dt, test_end_dt)
        self.test_interval.data = data_history
        self.control_interval = ChangePointInterval(control_start_dt, control_end_dt)
        self.control_interval.data = data_history

    def _init_data(self, data: Optional[TimeSeriesData]) -> None:
        self.data_history = data

    def _update_data(self, data: TimeSeriesData) -> None:
        """
        Updates the data with new data.
        """
        data_history = self.data_history
        assert data_history is not None
        self.data_history = TimeSeriesData(
            time=pd.concat([data_history.time, data.time], copy=False),
            value=pd.concat([data_history.value, data.value], copy=False),
        )

    def predict(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData] = None,
        **kwargs: Any,
    ) -> AnomalyResponse:
        """
        Predict is not implemented.
        """

        raise ValueError("predict is not implemented, call fit_predict() instead")


class MultiStatSigDetectorModel(StatSigDetectorModel):

    """
    MultiStatSigDetectorModel is a multivariate version of the StatSigDetector.  It applies a univariate
    t-test to each of the components of the multivariate time series to see if the means between the control
    and test periods are significantly different.  Then it uses a false discovery rate controlling procedure
    rate (FDR) controlling procedure (https://en.wikipedia.org/wiki/False_discovery_rate#Controlling_procedure)
    to adjust the p-values, reducing the noise the the alerts that are triggered by the detector.  The default
    FDR controlling procedure is the Benjamini-Hochberg procedure, but this can be adjusted when initializing
    the model.

    Like with the StatSigDetector, we start with the history data, and then as for the current data,
    we apply a rolling window, adding one data point at a time from the current data,
    and detecting significant change. The T-statistics we return here are based on the adjusted p-values
    from the FDR controlling procedure.

    We suggest using n_control >= 30 to get good estimates

    Attributes:
        n_control: int, number of data points(or time units) of history to compare with
        n_test: int, number of points(or time_units) to compare the history with
        serialized_model: Optional, serialized json containing the parameters
        time_units: str, units of time used to measure the intervals. If not provided
                    we infer it from the provided data
        method: str, indicates the FDR controlling method used for adjusting the p-values.
            Defaults to 'fdr_bh' for Benjamini-Hochberg.  Inputs for other FDR controlling methods
            can be found at https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.multipletests.html
        rem_season: default value is False, if remove seasonality for historical data and data
        seasonal_period: str, default value is 'weekly'. Other possible values: 'daily', 'biweekly', 'monthly', 'yearly'
        skip_rescaling: bool. If we'd like skip rescaling p-values for multivariate timeseires when calling Percentagechange class
        use_corrected_scores: bool, default value is False, using original t-scores or correct t-scores.
        min_perc_change: float, minimum percentage change, for a non zero score. Score will be clipped to zero if the absolute value of the percentage chenge is less than this value
    >>> # Example usage:
    >>> # history and ts_pt are TimeSeriesData objects and history is larger
    >>> # than (n_control + n_test) so that we have sufficient history to
    >>> # run the detector
    >>> n_control = 28
    >>> n_test = 7
    >>> import random
    >>> control_time = pd.date_range(start='2018-01-01', freq='D', periods=(n_control + n_test))
    >>> test_time = pd.date_range(start='2018-02-05', freq='D', periods=n_test)
    >>> num_seq = 5
    >>> control_val = [np.random.randn(len(control_time)) for _ in range(num_seq)]
    >>> test_val = [np.random.randn(len(test_time)) for _ in range(num_seq)]
    >>> hist_ts =
        TimeSeriesData(
            pd.DataFrame(
                {
                    **{"time": control_time},
                    **{f"ts_{i}": control_val[i] for i in range(num_seq)},
                }
            )
        )
    >>> data_ts =
        TimeSeriesData(
            pd.DataFrame(
                {
                    **{"time": test_time},
                    **{f"ts_{i}": test_val[i] for i in range(num_seq)},
                }
            )
        )
    >>> ss_detect = MultiStatSigDetectorModel(n_control=n_control, n_test=n_test)
    >>> anom = ss_detect.fit_predict(data=data_ts, historical_data=hist_ts)
    """

    def __init__(
        self,
        n_control: Optional[int] = None,
        n_test: Optional[int] = None,
        serialized_model: Optional[bytes] = None,
        time_unit: Optional[str] = None,
        rem_season: bool = False,
        seasonal_period: str = "weekly",
        method: str = "fdr_bh",
        skip_rescaling: bool = False,
        use_corrected_scores: bool = False,
        min_perc_change: float = 0.0,
    ) -> None:

        StatSigDetectorModel.__init__(
            self,
            n_control=n_control,
            n_test=n_test,
            serialized_model=serialized_model,
            time_unit=time_unit,
            rem_season=rem_season,
            seasonal_period=seasonal_period,
            min_perc_change=min_perc_change,
        )
        self.method = method
        self.skip_rescaling = skip_rescaling
        self.use_corrected_scores = use_corrected_scores

    def fit_predict(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData] = None,
        **kwargs: Any,
    ) -> AnomalyResponse:
        """
        This is the main working function.
        The function returns an AnomalyResponse object of length
        equal to the length of the data
        We require len(historical_data) > (n_control + n_test)

        Args:
            data: TimeSeriesData, A multivariate TimeSeriesData for which we are running the MultiStatSigDetectorModel
            historical_data: Optional[TimeSeriesData] Historical data used to do detection for initial points in data
        """

        if data.is_univariate():
            msg = "Input is univariate but MultiStatSigDetector expected multivariate input."
            logging.error(msg)
            raise ValueError(msg)

        self._set_time_unit(data=data, historical_data=historical_data)
        self.last_N = len(data)

        # this ensures we start with a default response of
        # the size of the data
        self._init_response(data)
        response = self.response
        assert response is not None
        # when there is no need to update
        # just return the initial response of zeros
        if not self._should_update(data=data, historical_data=historical_data):
            return response.get_last_n(self.last_N)

        # handle cases where there is either no historical  data, or
        # not enough historical data
        data, historical_data = self._handle_not_enough_history(
            data=data,
            historical_data=historical_data,
        )
        self.data = data

        # remove seasonality
        if self.rem_season:
            sh_data = SeasonalityHandler(data=data, seasonal_period=self.seasonal_period)

            self.data_season = sh_data.get_seasonality()
            data = sh_data.remove_seasonality()

            if historical_data:
                sh_hist_data = SeasonalityHandler(
                    data=historical_data,
                    seasonal_period=self.seasonal_period,
                )
                historical_data = sh_hist_data.remove_seasonality()

        # first initialize this with the historical data
        self._init_data(historical_data)
        self._init_control_test(data if historical_data is None else historical_data)
        # set the flag to true
        self.is_initialized = True

        # now run through the data to get the prediction

        for i in range(len(data)):
            ts_pt = TimeSeriesData(
                pd.DataFrame(
                    {
                        **{"time": data.time.iloc[i]},
                        **{c: data.value[c].iloc[i] for c in data.value.columns},
                    },
                    index=[0],
                    copy=False,
                )
            )
            self._update_data(ts_pt)
            self._update_control_test(ts_pt)
            self._update_response(ts_pt.time.iloc[0])

        # add seasonality back
        if self.rem_season:
            data_season = self.data_season
            confidence_band = response.confidence_band
            predicted_ts = response.predicted_ts
            assert data_season is not None
            assert confidence_band is not None
            assert predicted_ts is not None

            start_idx = len(response.scores) - len(data_season)
            datatime = response.scores.time

            self.response = AnomalyResponse(
                scores=response.scores,
                confidence_band=ConfidenceBand(
                    upper=TimeSeriesData(
                        time=datatime,
                        value=pd.concat(
                            [
                                pd.DataFrame(
                                    confidence_band.upper.value.values[:start_idx, :],
                                    copy=False,
                                ),
                                pd.DataFrame(
                                    np.asarray(confidence_band.upper.value.values[start_idx:, :])
                                    + np.asarray(data_season.value.values),
                                    copy=False,
                                ),
                            ],
                            copy=False,
                        ),
                    ),
                    lower=TimeSeriesData(
                        time=datatime,
                        value=pd.concat(
                            [
                                pd.DataFrame(
                                    confidence_band.lower.value.values[:start_idx, :],
                                    copy=False,
                                ),
                                pd.DataFrame(
                                    np.asarray(confidence_band.lower.value.values[start_idx:, :])
                                    + np.asarray(data_season.value.values),
                                    copy=False,
                                ),
                            ],
                            copy=False,
                        ),
                    ),
                ),
                predicted_ts=TimeSeriesData(
                    time=datatime,
                    value=pd.concat(
                        [
                            pd.DataFrame(predicted_ts.value.values[:start_idx, :], copy=False),
                            pd.DataFrame(
                                np.asarray(predicted_ts.value.values[start_idx:, :])
                                + np.asarray(data_season.value.values),
                                copy=False,
                            ),
                        ],
                        copy=False,
                    ),
                ),
                anomaly_magnitude_ts=response.anomaly_magnitude_ts,
                stat_sig_ts=response.stat_sig_ts,
            )
        assert self.response is not None
        return self.response.get_last_n(self.last_N)

    def _init_response(self, data: TimeSeriesData) -> None:

        zeros_df = pd.DataFrame(
            {
                **{"time": data.time},
                **{c: pd.Series(np.zeros(len(data))) for c in data.value.columns},
            },
            copy=False,
        )

        init_df = pd.DataFrame(
            {
                **{"time": data.time},
                **{c: data.value[c].values for c in data.value.columns},
            },
            copy=False,
        )

        self.response = AnomalyResponse(
            scores=TimeSeriesData(zeros_df.copy()),
            confidence_band=ConfidenceBand(
                upper=TimeSeriesData(init_df.copy()),
                lower=TimeSeriesData(init_df.copy()),
            ),
            predicted_ts=TimeSeriesData(init_df.copy()),
            anomaly_magnitude_ts=TimeSeriesData(zeros_df.copy()),
            stat_sig_ts=TimeSeriesData(zeros_df.copy()),
        )

    def _update_response(self, date: pd.Timestamp) -> None:
        """
        updates the current response with data from date
        """
        assert self.test_interval is not None
        assert self.control_interval is not None
        perc_change = PercentageChange(
            current=self.test_interval,
            previous=self.control_interval,
            method=self.method,
            skip_rescaling=self.skip_rescaling,
            use_corrected_scores=self.use_corrected_scores,
            min_perc_change=self.min_perc_change,
        )
        assert self.response is not None
        self.response.inplace_update(
            time=date,
            score=perc_change.score,
            ci_upper=perc_change.ci_upper,
            ci_lower=perc_change.ci_lower,
            pred=perc_change.mean_previous,
            anom_mag=perc_change.mean_difference,
            stat_sig=perc_change.stat_sig,
        )

    def _init_control_test(self, data: TimeSeriesData) -> None:
        """
        initializes the control and test intervals
        """
        (
            control_start_dt,
            control_end_dt,
            test_start_dt,
            test_end_dt,
        ) = self._get_start_end_dates(data)
        data_history = self.data_history
        assert data_history is not None

        self.test_interval = ChangePointInterval(test_start_dt, test_end_dt)
        self.test_interval.data = data_history
        self.control_interval = ChangePointInterval(control_start_dt, control_end_dt)
        self.control_interval.data = data_history

    def _update_control_test(self, data: TimeSeriesData) -> None:
        """
        updates control and test with new data
        """
        (
            control_start_dt,
            control_end_dt,
            test_start_dt,
            test_end_dt,
        ) = self._get_start_end_dates(data)
        data_history = self.data_history
        assert data_history is not None

        self.test_interval = ChangePointInterval(test_start_dt, test_end_dt)
        self.test_interval.data = data_history
        self.control_interval = ChangePointInterval(control_start_dt, control_end_dt)
        self.control_interval.data = data_history


class SeasonalityHandler:
    """
    SeasonalityHandler is a class that do timeseries STL decomposition for detecors
    Attributes:
        data: TimeSeriesData that need to be decomposed
        seasonal_period: str, default value is 'weekly'. Other possible values: 'daily', 'biweekly', 'monthly', 'yearly'

    >>> # Example usage:
    >>> from kats.utils.simulator import Simulator
    >>> sim = Simulator(n=120, start='2018-01-01')
    >>> ts = sim.level_shift_sim(cp_arr = [60], level_arr=[1.35, 1.05], noise=0.05, seasonal_period=7, seasonal_magnitude=0.575)
    >>> sh = SeasonalityHandler(data=ts, seasonal_period='weekly')
    >>> sh.get_seasonality()
    >>> sh.remove_seasonality()
    """

    def __init__(
        self,
        data: TimeSeriesData,
        seasonal_period: str = "weekly",
        **kwargs: Any,
    ) -> None:
        self.data = data

        _map = {"daily": 1, "weekly": 7, "biweekly": 14, "monthly": 30, "yearly": 365}
        if seasonal_period not in _map:
            msg = "Invalid seasonal_period, possible values are 'daily', 'weekly', 'biweekly', 'monthly', and 'yearly'"
            logging.error(msg)
            raise ValueError(msg)
        self.seasonal_period: int = _map[seasonal_period] * 24  # change to hours

        self.low_pass_jump_factor: float = kwargs.get("lpj_factor", 0.15)
        self.trend_jump_factor: float = kwargs.get("tj_factor", 0.15)

        self.frequency: pd.Timedelta = self.data.infer_freq_robust()
        self.frequency_sec: int = int(self.frequency.total_seconds())
        self.frequency_sec_str: str = str(self.frequency_sec) + "s"

        # calculate resample base in second level
        time0 = pd.to_datetime(self.data.time[0])
        # calculate remainder as resampling base
        resample_base_sec = (
            time0.day * 24 * 60 * 60 + time0.hour * 60 * 60 + time0.minute * 60 + time0.second
        ) % self.frequency_sec

        self.decomposer_input: TimeSeriesData = self.data.interpolate(
            freq=self.frequency_sec_str,
            base=resample_base_sec,
        )

        data_time_idx = self.decomposer_input.time.isin(self.data.time)
        if len(self.decomposer_input.time[data_time_idx]) != len(self.data):
            raise ValueError(IRREGULAR_GRANULARITY_ERROR)

        self.period = int(self.seasonal_period * 60 * 60 / self.frequency.total_seconds())

        if (
            int((self.period + 1) * self.low_pass_jump_factor) < 1
            or int((self.period + 1) * self.trend_jump_factor) < 1
        ):
            msg = (
                "Invalid low_pass_jump or trend_jump_factor value, try a larger factor or try a larger seasonal_period"
            )
            logging.error(msg)
            raise ValueError(msg)

        self.decomp: Optional[dict[str, Any]] = None

        self.ifmulti: bool = False
        # for multi-variate TS
        if len(self.data.value.values.shape) != 1:
            self.ifmulti = True
            self.num_seq: int = self.data.value.values.shape[1]

        self.data_season = TimeSeriesData(time=self.data.time, value=self.data.value)
        self.data_nonseason = TimeSeriesData(time=self.data.time, value=self.data.value)

    def _decompose(self) -> None:
        if not self.ifmulti:
            decomposer = TimeSeriesDecomposition(
                self.decomposer_input,
                period=self.period,
                robust=True,
                seasonal_deg=0,
                trend_deg=1,
                low_pass_deg=1,
                low_pass_jump=int((self.period + 1) * self.low_pass_jump_factor),
                seasonal_jump=1,
                trend_jump=int((self.period + 1) * self.trend_jump_factor),
            )

            self.decomp = decomposer.decomposer()
            return

        self._decompose_multi()

    def _decompose_multi(self) -> None:
        self.decomp = {}
        for i in range(self.num_seq):
            temp_ts = TimeSeriesData(
                time=self.decomposer_input.time,
                value=pd.Series(self.decomposer_input.value.values[:, i], copy=False),
            )
            decomposer = TimeSeriesDecomposition(
                temp_ts,
                period=self.period,
                robust=True,
                seasonal_deg=0,
                trend_deg=1,
                low_pass_deg=1,
                low_pass_jump=int((self.period + 1) * self.low_pass_jump_factor),
                seasonal_jump=1,
                trend_jump=int((self.period + 1) * self.trend_jump_factor),
            )
            assert self.decomp is not None
            self.decomp[str(i)] = decomposer.decomposer()

    def remove_seasonality(self) -> TimeSeriesData:
        if self.decomp is None:
            self._decompose()
        if not self.ifmulti:
            decomp = self.decomp
            assert decomp is not None
            data_time_idx = decomp["rem"].time.isin(self.data_nonseason.time)

            self.data_nonseason.value = pd.Series(
                decomp["rem"][data_time_idx].value + decomp["trend"][data_time_idx].value,
                name=self.data_nonseason.value.name,
                copy=False,
            )
            return self.data_nonseason
        decomp = self.decomp
        assert decomp is not None
        data_time_idx = decomp[str(0)]["rem"].time.isin(self.data_nonseason.time)
        for i in range(self.num_seq):
            self.data_nonseason.value.iloc[:, i] = pd.Series(
                decomp[str(i)]["rem"][data_time_idx].value + decomp[str(i)]["trend"][data_time_idx].value,
                name=self.data_nonseason.value.iloc[:, i].name,
                copy=False,
            )

        return self.data_nonseason

    def get_seasonality(self) -> TimeSeriesData:

        if self.decomp is None:
            self._decompose()
        decomp = self.decomp
        assert decomp is not None
        if not self.ifmulti:
            data_time_idx = decomp["seasonal"].time.isin(self.data_season.time)
            self.data_season.value = pd.Series(
                decomp["seasonal"][data_time_idx].value,
                name=self.data_season.value.name,
                copy=False,
            )
            return self.data_season

        data_time_idx = decomp[str(0)]["seasonal"].time.isin(self.data_season.time)
        for i in range(self.num_seq):
            self.data_season.value.iloc[:, i] = pd.Series(
                decomp[str(i)]["seasonal"][data_time_idx].value,
                name=self.data_season.value.iloc[:, i].name,
                copy=False,
            )

        return self.data_season



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from . import cusum_detection  # noqa # usort: skip
from . import cusum_model  # noqa # usort: skip
from . import detector  # noqa # usort: skip
from . import detector_consts  # noqa # usort: skip
from . import hourly_ratio_detection  # noqa # usort: skip
from . import outlier  # noqa # usort: skip
from . import residual_translation  # noqa # usort: skip
from . import robust_stat_detection  # noqa # usort: skip
from . import seasonality  # noqa # usort: skip
from . import stat_sig_detector  # noqa # usort: skip
from . import trend_mk  # noqa # usort: skip
from .meta_learning import metalearning_detection_model  # noqa # usort: skip



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""Base I/O code for time series data in Kats

This is a base implementation to load datasets for test and evaluation/benchmarking
purposes. We currently support the following data sets:

1. air passengers data
2. m3 meta data
3. Peyton manning data
4. retail sales data
5. yosemite temps data
6. multi ts data
7. mean change detection test data
8. multivariate anomaly simulated data
"""

from typing import List
from typing import Tuple

import numpy as np
import pandas as pd
from scipy.special import expit  # @manual

from kats.consts import TimeSeriesData


def gen_no_trend_data_ndim(time: pd.Series, ndim: int = 1) -> TimeSeriesData:
    np.random.seed(20)
    n_days = len(time)
    data = np.ones((n_days, ndim)) * np.random.randint(1000, size=(1, ndim))
    no_trend_data = pd.DataFrame(data)
    no_trend_data["time"] = time

    return TimeSeriesData(no_trend_data)


def gen_trend_data_ndim(
    time: pd.Series,
    seasonality: float = 0.00,
    change_smoothness: float = 5.0,
    ndim: int = 1,
) -> Tuple[TimeSeriesData, List[float]]:
    np.random.seed(20)

    n_days = len(time)
    ix = np.array([np.arange(n_days) for i in range(ndim)])
    initial = np.random.randint(9000.0, 10000.0, size=(ndim, 1))
    trend_change = -np.random.randint(60, size=(ndim, 1))
    trend = np.random.randint(2.0, 6.0, size=(ndim, 1))
    noise = np.array([1e-3] * ndim).reshape((ndim, 1))
    t_change = np.random.randint(int(0.4 * n_days), int(0.7 * n_days), size=(ndim, 1))

    data = (
        (initial + trend * ix + trend_change * (ix - t_change) * expit((ix - t_change)))
        * (1 - seasonality * (ix % 7 >= 5))
        * np.array([np.cumprod(1 + noise[i] * np.random.randn(n_days)) for i in range(ndim)])
    )

    trend_data = pd.DataFrame(data.T)
    trend_data["time"] = time

    t_change = [t_change[i][0] for i in range(len(t_change))]

    return TimeSeriesData(trend_data), t_change



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import unittest
from typing import Dict
from typing import Union
from typing import cast
from unittest import TestCase

import pandas as pd
from parameterized.parameterized import parameterized

from kats.compat.pandas import assert_frame_equal
from kats.consts import TimeSeriesData
from kats.data.utils import load_data
from kats.models.linear_model import LinearModel
from kats.models.linear_model import LinearModelParams
from kats.tests.models.test_models_dummy_data import AIR_FCST_LINEAR_95
from kats.tests.models.test_models_dummy_data import AIR_FCST_LINEAR_99
from kats.tests.models.test_models_dummy_data import PEYTON_FCST_LINEAR_95
from kats.tests.models.test_models_dummy_data import PEYTON_FCST_LINEAR_99
from kats.tests.models.test_models_dummy_data import PEYTON_FCST_LINEAR_INVALID_NEG_ONE
from kats.tests.models.test_models_dummy_data import PEYTON_FCST_LINEAR_INVALID_ZERO
from kats.tests.models.test_models_dummy_data import PEYTON_FCST_LINEAR_NAN
from kats.tests.models.test_models_dummy_data import PEYTON_INPUT_NAN

TEST_DATA: Dict[
    str,
    Union[
        Dict[str, Union[TimeSeriesData, int, pd.DataFrame, str]],
        Dict[str, Union[TimeSeriesData, pd.DataFrame, str]],
        Dict[str, TimeSeriesData],
    ],
] = {
    "daily": {
        "ts": TimeSeriesData(load_data("peyton_manning.csv", reset_columns=True)),
        "ts_nan": TimeSeriesData(df=PEYTON_INPUT_NAN),
        "freq": "D",
        "res_95": PEYTON_FCST_LINEAR_95,
        "res_99": PEYTON_FCST_LINEAR_99,
        "res_invalid": PEYTON_FCST_LINEAR_NAN,
        "res_invalid_param_1": PEYTON_FCST_LINEAR_INVALID_ZERO,
        "res_invalid_param_2": PEYTON_FCST_LINEAR_INVALID_NEG_ONE,
        "invalid_param_1": 0,
        "invalid_param_2": -1,
    },
    "monthly": {
        "ts": TimeSeriesData(load_data("air_passengers.csv", reset_columns=True)),
        "freq": "MS",
        "res_95": AIR_FCST_LINEAR_95,
        "res_99": AIR_FCST_LINEAR_99,
    },
    "multi": {
        "ts": TimeSeriesData(load_data("multivariate_anomaly_simulated_data.csv")),
    },
}


class LinearModelTest(TestCase):
    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            [
                "monthly",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["freq"],
                TEST_DATA["monthly"]["res_95"],
                TEST_DATA["monthly"]["res_99"],
            ],
            [
                "daily",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["freq"],
                TEST_DATA["daily"]["res_95"],
                TEST_DATA["daily"]["res_99"],
            ],
            [
                "invalid",
                TEST_DATA["daily"]["ts_nan"],
                TEST_DATA["daily"]["freq"],
                TEST_DATA["daily"]["res_invalid"],
                TEST_DATA["daily"]["res_invalid"],
            ],
        ]
    )
    def test_fcst(
        self,
        name: str,
        ts: TimeSeriesData,
        freq: str,
        truth_95: pd.DataFrame,
        truth_99: pd.DataFrame,
    ) -> None:
        # Set up params
        params_95 = LinearModelParams(alpha=0.05)
        params_99 = LinearModelParams(alpha=0.01)
        params_95.validate_params()
        params_99.validate_params()
        # Fit forecast
        m_95 = LinearModel(ts, params_95)
        m_99 = LinearModel(ts, params_99)
        m_95.fit()
        m_99.fit()
        res_95 = m_95.predict(steps=30, freq=freq)
        res_99 = m_99.predict(steps=30, freq=freq)
        # Test result
        assert_frame_equal(truth_95, res_95)
        assert_frame_equal(truth_99, res_99)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            [
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["invalid_param_1"],
                TEST_DATA["daily"]["freq"],
                TEST_DATA["daily"]["res_invalid_param_1"],
            ],
            [
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["invalid_param_2"],
                TEST_DATA["daily"]["freq"],
                TEST_DATA["daily"]["res_invalid_param_2"],
            ],
        ]
    )
    def test_invalid_params(self, ts: TimeSeriesData, invalid_param: float, freq: str, truth: pd.DataFrame) -> None:
        # Set up params
        params = LinearModelParams(alpha=invalid_param)
        # Fit forecast
        m = LinearModel(ts, params)
        m.fit()
        res = m.predict(steps=30, freq=freq, include_history=True)
        # Test result
        assert_frame_equal(truth, res)

    def test_multivar(self) -> None:
        self.assertRaises(ValueError, LinearModel, TEST_DATA["multi"]["ts"], None)

    def test_exec_plot(self) -> None:
        # Set up params
        params = LinearModelParams(alpha=0.05)
        # Fit forecast
        m = LinearModel(cast(TimeSeriesData, TEST_DATA["daily"]["ts"]), params)
        m.fit()
        _ = m.predict(steps=2, freq=TEST_DATA["daily"]["freq"])
        # Test plotting
        m.plot()

    def test_name(self) -> None:
        m = LinearModel(cast(TimeSeriesData, TEST_DATA["daily"]["ts"]), LinearModelParams())
        self.assertEqual(m.__str__(), "Linear Model")

    def test_search_space(self) -> None:
        m = LinearModel(cast(TimeSeriesData, TEST_DATA["daily"]["ts"]), LinearModelParams())
        self.assertEqual(
            m.get_parameter_search_space(),
            [
                {
                    "name": "alpha",
                    "type": "choice",
                    "value_type": "float",
                    "values": [0.01, 0.05, 0.1, 0.25],
                    "is_ordered": True,
                },
            ],
        )


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
 Bayesian estimation of Vector Autoregressive Model using
 Minnesota prior on the coefficient matrix. This version is
 useful for regularization when they are too many coefficients
 to be estimated.

 Implementation inspired by the following two articles/papers:
    https://www.mathworks.com/help/econ/normalbvarm.html#mw_4a1ab118-9ef3-4380-8c5a-12b848254117
    http://apps.eui.eu/Personal/Canova/Articles/ch10.pdf (page 5)
"""

import logging
from dataclasses import dataclass
from datetime import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from numpy.linalg import inv
from scipy.linalg import block_diag

import kats.models.model as m
from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.consts import _log_error


@dataclass
class BayesianVARParams(Params):
    """Parameter class for Bayesian VAR model

    Attributes:
        p: Historical lag to use
        Below parameters are hyperparameters in the covariance matrix for
        coefficient prior. See page 5 in
        http://apps.eui.eu/Personal/Canova/Articles/ch10.pdf for more details.
        phi_0: tightness on the variance of the first lag
        phi_1: relative tightness of other variables
        phi_2: relative tightness of the exogenous variables
        phi_3: decay with lag is parameterized as lag^phi_3
    """

    p: int = 5
    phi_0: float = 0.02
    phi_1: float = 0.25
    phi_2: float = 20
    phi_3: float = 3

    def validate_params(self) -> None:
        if self.p <= 0:
            raise ValueError(f"Lag order must be positive, but got {self.p}")
        if self.phi_0 <= 0:
            raise ValueError(f"phi_0 must be positive, but got {self.phi_0}")
        if self.phi_1 <= 0 or self.phi_1 > 1:
            raise ValueError(
                f"phi_1 must be positive and at most 1, but got {self.phi_1}. "
                "See page 5 of http://apps.eui.eu/Personal/Canova/Articles/ch10.pdf"
                "for more details."
            )
        if self.phi_2 <= 0:
            raise ValueError(f"phi_2 must be positive, but got {self.phi_2}")
        if self.phi_3 <= 0:
            raise ValueError(f"phi_3 must be positive, but got {self.phi_3}")


class BayesianVAR(m.Model[BayesianVARParams]):
    """
    Model class for bayesian VAR

    This class provides fit, predict, and plot methods for bayesian VAR model

    Attributes:
        data: the input time series data as `TimeSeriesData`
        params: the parameter class defined with `BayesianVARParams`
    """

    sigma_ols: Optional[np.ndarray] = None
    v_posterior: Optional[np.ndarray] = None
    mu_posterior: Optional[np.ndarray] = None
    resid: Optional[pd.DataFrame] = None
    forecast: Optional[Dict[str, TimeSeriesData]] = None
    forecast_max_time: Optional[datetime] = None
    data: TimeSeriesData
    time_freq: str
    X: np.ndarray
    Y: np.ndarray
    m: int
    T: int
    r: int
    p: int
    phi_0: float
    phi_1: float
    phi_2: float
    phi_3: float
    N: int
    num_mu_coefficients: int
    fitted: bool = False
    forecast_vals: Optional[List[np.ndarray]] = None

    def __init__(self, data: TimeSeriesData, params: BayesianVARParams) -> None:
        if data.is_univariate():
            msg = "Bayesian VAR Model only accepts multivariate time series."
            raise _log_error(msg)

        self.data = data
        self.time_freq = BayesianVAR._check_get_freq(data)
        self.X, self.Y = BayesianVAR._convert_timeseries_np(data)
        assert self.X.shape[1] == self.Y.shape[1], "Expected same amount of data on time axis for X and Y"

        self.m, self.T = self.Y.shape
        self.r = self.X.shape[0]
        self.p = params.p
        self.phi_0 = params.phi_0
        self.phi_1 = params.phi_1
        self.phi_2 = params.phi_2
        self.phi_3 = params.phi_3
        self.N = (self.m * self.p) + self.r + 1
        self.num_mu_coefficients = self.m * self.N

        logging.info(f"Initializing Bayesian VAR model with: {self}")

    def __str__(self) -> str:
        return (
            f"BayesianVAR(p={self.p}, m={self.m}, r={self.r}, T={self.T}, "
            f"N={self.N}, phi_0={self.phi_0}, phi_1={self.phi_1}, "
            f"phi_2={self.phi_2}, phi_3={self.phi_3})"
        )

    @staticmethod
    def _check_get_freq(data: TimeSeriesData) -> str:
        """Checks for consistent time frequency."""
        freq = pd.infer_freq(data.time)
        if freq is None:
            raise ValueError(
                "Unable to infer time series frequency. Please check for "
                "missing or duplicate times or irregularly-spaced times."
            )

        return freq

    @staticmethod
    def _convert_timeseries_np(
        timeseries: TimeSeriesData,
    ) -> Tuple[np.ndarray, np.ndarray]:
        data_df = timeseries.to_dataframe()
        Y = data_df.drop(columns=[timeseries.time_col_name]).to_numpy().T
        X = np.expand_dims(pd.RangeIndex(0, len(timeseries)), axis=0)
        return X, Y

    def _get_training_residuals(self) -> pd.DataFrame:
        residuals = []

        # create dataframe with each column corresponding to the residual
        p = self.p
        T = self.T
        times = self.data.time.iloc[p:T]
        self.forecast_vals = forecast_vals = []
        if times.empty:
            logging.info(
                "Performing one-step ahead forecasting on history from step "
                f"{self.p} to {self.T-1} (t={times[0]} to {times[-1]}) inclusive."
            )
        for t in range(p, T):
            point_pred = self._evaluate_point_t(self.X, self.Y, t)
            forecast_vals.append(point_pred)
            residuals.append(self.Y[:, t] - point_pred)
        df_resid = pd.DataFrame(residuals, index=times, columns=self.data.value.columns, copy=False)

        return df_resid

    def fit(self) -> None:
        """Fit Bayesian VAR model"""

        self.sigma_ols = self._compute_sigma_ols()

        mu_prior = np.zeros((self.m, self.N))
        for i in range(self.m):
            mu_prior[i, self.p * i] = 1
        mu_prior = mu_prior.flatten()

        v_prior = self._construct_v_prior()

        Z_sig_Z_sum = 0
        Z_sig_y_sum = 0

        num_mu = self.num_mu_coefficients
        for t in range(self.p, self.T):
            Z_t = self._construct_Zt(self.X, self.Y, t)  # shape: m x [m * (m * p + r + 1)]

            z_sum_term = (Z_t.T @ inv(self.sigma_ols)) @ Z_t  # shape: [m * (m * p + r + 1)] x [m * (m * p + r + 1)]
            y_sum_term = (Z_t.T @ inv(self.sigma_ols)) @ self.Y[:, t]  # shape: [m * (m * p + r + 1)] x 1

            assert (
                num_mu,
                num_mu,
            ) == z_sum_term.shape, f"Expected {(num_mu, num_mu)}, got {z_sum_term.shape}"
            assert (num_mu,) == y_sum_term.shape, f"Expected {(num_mu,)}, got {y_sum_term.shape}"

            Z_sig_Z_sum += z_sum_term
            Z_sig_y_sum += y_sum_term

        v_posterior = inv(inv(v_prior) + Z_sig_Z_sum)  # shape: [m * (m * p + r + 1)] x [m * (m * p + r + 1)]
        self.v_posterior = v_posterior
        assert (
            num_mu,
            num_mu,
        ) == v_posterior.shape, f"Expected {(num_mu, num_mu)}, got {v_posterior.shape}"

        mu_posterior = v_posterior @ (inv(v_prior) @ mu_prior + Z_sig_y_sum)  # shape: [m * (m * p + r + 1)] x 1
        self.mu_posterior = mu_posterior
        assert (num_mu,) == mu_posterior.shape, f"Expected {(num_mu,)}, got {mu_posterior.shape}"
        self.resid = self._get_training_residuals()
        self.fitted = True

    def _construct_z(self, X: np.ndarray, Y: np.ndarray, t: int) -> np.ndarray:
        assert t >= self.p, f"Need t={t} >= p={self.p}."
        assert self.r == X.shape[0]
        assert self.m == Y.shape[0]

        new_yt = np.fliplr(Y[:, t - self.p : t]).flatten()
        z = np.concatenate([new_yt, X[:, t].T, np.array([1])], axis=0)  # shape: [(m * p + r + 1) x 1]

        assert (self.N,) == z.shape, f"Expected {(self.N,)} but got {z.shape}"

        return z

    def _construct_Zt(self, X: np.ndarray, Y: np.ndarray, t: int) -> np.ndarray:
        z = self._construct_z(X, Y, t)
        Z_t = block_diag(*([z] * self.m))

        assert (
            self.m,
            self.num_mu_coefficients,
        ) == Z_t.shape, f"Expected {(self.m, self.num_mu_coefficients)}, got {Z_t.shape}"

        return Z_t  # shape: m x [m * (m * p + m + 1)]

    def _construct_X_OLS(self) -> np.ndarray:
        X_OLS = np.zeros((self.N, self.T - self.p))

        for t in range(self.p, self.T):
            X_OLS[:, t - self.p] = self._construct_z(self.X, self.Y, t)  # X_OLS ignores first p values

        return X_OLS

    def _compute_sigma_ols(self) -> np.ndarray:
        Y_suffix = self.Y[:, self.p :]
        X_OLS = self._construct_X_OLS()

        beta_ols = (Y_suffix @ X_OLS.T) @ inv(X_OLS @ X_OLS.T)
        sse = (Y_suffix - beta_ols @ X_OLS) @ (Y_suffix - beta_ols @ X_OLS).T  # should produce [m x m] matrix

        assert (
            self.m,
            self.m,
        ) == sse.shape, f"Expected {(self.m, self.m)}, but got {sse.shape}"
        assert self.T > (self.m * self.p) + 1

        return sse / float(self.T - (self.m * self.p) - 1)

    def _sigma_ijl(
        self,
        i: int,
        j: Optional[int],
        lag: Optional[int],
        variance: np.ndarray,
        is_exogenous: bool,
    ) -> float:
        """
        Taken from page 5 of http://apps.eui.eu/Personal/Canova/Articles/ch10.pdf
        """

        def h(x: float) -> float:
            return x**self.phi_3

        if i == j:
            assert lag is not None
            return self.phi_0 / h(lag)
        elif is_exogenous:
            return self.phi_0 * self.phi_2
        else:  # endogenous variable j
            assert lag is not None
            return self.phi_0 * (self.phi_1 / h(lag)) * (variance[j] / variance[i])

    def _construct_v_prior(self) -> np.ndarray:
        num_mu = self.num_mu_coefficients
        cov = np.zeros((num_mu, num_mu))

        variance = np.var(self.Y, axis=1)

        element_ind = 0

        for i in range(self.m):
            for j in range(self.m):  # iterate through the m classes of lagged variables
                for lag in range(1, self.p + 1):  # iterate through the lags
                    cov[element_ind][element_ind] = self._sigma_ijl(i, j, lag, variance, is_exogenous=False)
                    element_ind += 1

            for _ex in range(self.r):  # exogenous variables
                cov[element_ind][element_ind] = self._sigma_ijl(i, None, None, variance, is_exogenous=True)
                element_ind += 1

            # constant term of 1
            cov[element_ind][element_ind] = self._sigma_ijl(i, None, None, variance, is_exogenous=True)
            element_ind += 1

        assert element_ind == num_mu, f"Final element: {element_ind}, expected: {num_mu}"

        return cov  # shape: [m * (m * p + r + 1)] x [m * (m * p + r + 1)] matrix

    def _evaluate_point_t(self, X_new: np.ndarray, Y_new: np.ndarray, t: int) -> np.ndarray:
        assert t >= self.p, f"Need t={t} > p={self.p}."

        Z_t = self._construct_Zt(X_new, Y_new, t)
        point_prediction = Z_t @ self.mu_posterior  # shape [m x 1]

        assert (self.m,) == point_prediction.shape

        return point_prediction

    def _look_ahead_step(self, X_ahead: np.ndarray, Y_curr: np.ndarray) -> np.ndarray:
        # Y_curr has one less element than X_ahead
        assert Y_curr.shape[1] + 1 == X_ahead.shape[1]
        t_ahead = X_ahead.shape[1] - 1  # -1 for 0-indexed array

        Z_t = self._construct_Zt(X_ahead, Y_curr, t_ahead)
        look_ahead_pred = Z_t @ self.mu_posterior  # shape [m x 1]

        assert (self.m,) == look_ahead_pred.shape

        return look_ahead_pred

    # pyre-fixme[14]: `predict` overrides method defined in `Model` inconsistently.
    # pyre-fixme[15]: `predict` overrides method defined in `Model` inconsistently.
    def predict(self, steps: int, include_history: bool = False, verbose: bool = False) -> Dict[str, TimeSeriesData]:
        """Predict with the fitted VAR model.

        Args:
            steps: Number of time steps to forecast
            include_history: return fitted values also

        Returns:
            Dictionary of predicted results for each metric. Each metric result
            has following columns: `time`, `fcst`, `fcst_lower`, and `fcst_upper`
            Note confidence intervals of forecast are not yet implemented.
        """
        if not self.fitted:
            raise ValueError("Must call fit() before predict().")
        if not steps and not include_history:
            raise ValueError("Forecast produced no values. Please set steps > 0 or " "include_history=True.")

        X_ahead = self.X
        Y_curr = self.Y
        T = self.T

        if include_history:
            times = self.data.time.iloc[self.p : T].tolist()
            forecast_vals = self.forecast_vals
            assert forecast_vals is not None
        else:
            times = []
            forecast_vals = []

        if steps:
            # future forecasting -- X_ahead is one time step ahead of Y_curr
            ahead_times = pd.date_range(start=self.data.time.iloc[-1], periods=steps + 1, freq=self.time_freq)[1:]
            logging.info(
                f"Performing future forecasting from step {T} to {T+steps-1} ("
                f"t={ahead_times[0]} to t={ahead_times[-1]}) inclusive."
            )
            assert len(ahead_times) == steps

            ahead_time = X_ahead[np.newaxis, :, -1]
            for step, time in zip(range(T, T + steps), ahead_times):
                X_ahead = np.concatenate([X_ahead, ahead_time + step], axis=1)
                look_ahead_pred = self._look_ahead_step(X_ahead, Y_curr)

                if verbose:
                    logging.info(f"Performing future forecasting at t={time}, step={step}.")

                forecast_vals.append(look_ahead_pred)

                Y_curr = np.concatenate([Y_curr, look_ahead_pred[:, np.newaxis]], axis=1)

            times += ahead_times

        forecast_length = len(times)

        assert forecast_length == len(forecast_vals), f"{forecast_length} != {len(forecast_vals)}"

        self.forecast = indiv_forecasts = {}
        self.forecast_max_time = times[-1]

        logging.warning(
            "Upper and lower confidence intervals of forecast not yet implemented " "for Bayesian VAR model."
        )

        for i, c in enumerate(self.data.value.columns.tolist()):
            c_forecast = pd.DataFrame(
                {
                    "time": times,
                    "fcst": [forecast_vals[f_t][i] for f_t in range(forecast_length)],
                    "fcst_lower": [-1] * forecast_length,
                    "fcst_upper": [-1] * forecast_length,
                },
                copy=False,
            )
            indiv_forecasts[c] = TimeSeriesData(c_forecast)

        return indiv_forecasts

    def plot(
        self,
        ax: Optional[plt.Axes] = None,
        figsize: Optional[Tuple[int, int]] = None,
        title: Optional[str] = "Input Timeseries & Forecast",
        ls: str = "--",
        **kwargs: Any,
    ) -> plt.Axes:
        """Plot forecasted results from Bayesian VAR model"""
        forecast = self.forecast
        if forecast is None:
            raise ValueError("Must call predict() before plot()")
        data = self.data

        if ax is None:
            if figsize is None:
                figsize = (20, 6)
            _, ax = plt.subplots(figsize=figsize)

        ax.set_title(title)

        for i, c in enumerate(data.value.columns):
            color = f"C{i}"
            ax.plot(data.time, data.value[c], c=color)
            ax.plot(forecast[c].time, forecast[c].value, ls, c=color)

        return ax

    @property
    def sigma_u(self) -> pd.DataFrame:
        return pd.DataFrame(
            self.sigma_ols,
            index=self.data.value.columns,
            columns=self.data.value.columns,
            copy=False,
        )

    @property
    def k_ar(self) -> int:
        return self.p



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""Base I/O code for time series data in Kats

This is a base implementation to load datasets for test and evaluation/benchmarking
purposes. We currently support the following data sets:

1. air passengers data
2. m3 meta data
3. Peyton manning data
4. retail sales data
5. yosemite temps data
6. multi ts data
7. mean change detection test data
8. multivariate anomaly simulated data
"""

import io
import os
import pkgutil
import sys
from typing import Union
from typing import overload

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

import pandas as pd

from kats.consts import TimeSeriesData


def load_data(file_name: str, reset_columns: bool = False) -> pd.DataFrame:
    """load data for tests and tutorial notebooks"""
    ROOT = "kats"
    if "kats" in os.getcwd().lower():
        path = "data/"
    else:
        path = "kats/data/"
    data_object = pkgutil.get_data(ROOT, path + file_name)
    assert data_object is not None
    df = pd.read_csv(io.BytesIO(data_object), encoding="utf8")
    if reset_columns:
        df.columns = ["time", "y"]
    return df


@overload
def load_air_passengers(return_ts: Literal[True]) -> TimeSeriesData:
    ...


@overload
def load_air_passengers(return_ts: Literal[False] = ...) -> pd.DataFrame:
    ...


def load_air_passengers(return_ts: bool = True) -> Union[pd.DataFrame, TimeSeriesData]:
    """Load and return air passengers time series dataset

    ==============================
    Length                     144
    Granularity              daily
    Label                     none

    Args:
        return_ts: return class:`kats.consts.TimeSeriesData` by default
                   return `pandas.DataFrame` otherwise

    Returns:
        data: class:`kats.consts.TimeSeriesData`
        file_name: `str`, the physical path of air passengers data set
        descr: `str`, full description of this data set

    Examples
    >>> from kats.data import load_air_passengers
    >>> air_passengers_ts = load_air_passengers()
    >>> print(air_passengers_ts)
              time    y
        0   1949-01-01  112
        1   1949-02-01  118
        2   1949-03-01  132
        3   1949-04-01  129
        4   1949-05-01  121
        ..         ...  ...
        139 1960-08-01  606
        140 1960-09-01  508
        141 1960-10-01  461
        142 1960-11-01  390
        143 1960-12-01  432

        [144 rows x 2 columns]
    """
    df = load_data("air_passengers.csv")
    df.columns = ["time", "y"]

    if return_ts:
        return TimeSeriesData(df)
    else:
        return df



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from typing import Any
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Union
from unittest import TestCase

import numpy as np
import pandas as pd
from parameterized.parameterized import parameterized

from kats.consts import TimeSeriesData
from kats.utils.datapartition import RollingOriginDataParition
from kats.utils.datapartition import SimpleDataPartition
from kats.utils.datapartition import SimpleTimestampDataPartition

DataPartition = Union[
    List[TimeSeriesData],
    Dict[Union[int, str], TimeSeriesData],
    TimeSeriesData,
]


class TrainTestData(NamedTuple):
    train: DataPartition
    test: DataPartition


DF: pd.DataFrame = pd.DataFrame({"time": pd.date_range("2021-05-06", periods=10), "value": np.arange(10)})
TS: TimeSeriesData = TimeSeriesData(DF)
TS_DICT: Dict[Union[int, str], TimeSeriesData] = {0: TS, 1: TS}
TS_LIST: List[TimeSeriesData] = [TS, TS]

ALL_DATASETS: List[DataPartition] = [TS, TS_DICT, TS_LIST]
# ground-truth for simple data partition

RAW_SDP_RES: List[List[TrainTestData]] = [
    [TrainTestData(train=TS[:9], test=TS[9:])],
    [TrainTestData(train={0: TS[:9], 1: TS[:9]}, test={0: TS[9:], 1: TS[9:]})],
    [TrainTestData(train=[TS[:9], TS[:9]], test=[TS[9:], TS[9:]])],
]

RAW_STDP_RES: List[List[TrainTestData]] = [
    [TrainTestData(train=TS[:5], test=TS[7:])],
    [TrainTestData(train={0: TimeSeriesData(), 1: TimeSeriesData()}, test={0: TS, 1: TS})],
    [TrainTestData(train=[TimeSeriesData(), TimeSeriesData()], test=[TS[7:], TS[7:]])],
]

RAW_RODP_RES: List[List[TrainTestData]] = [
    [
        TrainTestData(train=TS[:5], test=TS[5:7]),
        TrainTestData(train=TS[:8], test=TS[8:]),
    ],
    [TrainTestData(train={0: TS[:5], 1: TS[:5]}, test={0: TS[5:], 1: TS[5:]})],
    [
        TrainTestData(train=[TS[:3]] * 2, test=[TimeSeriesData()] * 2),
        TrainTestData(train=[TS[3:6]] * 2, test=[TimeSeriesData()] * 2),
        TrainTestData(train=[TS[7:10]] * 2, test=[TimeSeriesData()] * 2),
    ],
    [
        TrainTestData(train=TS[:3], test=TS[4:6]),
        TrainTestData(train=TS[2:5], test=TS[6:8]),
        TrainTestData(train=TS[4:7], test=TS[8:10]),
    ],
]


class TestSimpleDataPartition(TestCase):
    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            [{"train_frac": 0.9, "test_frac": 0.1}],
            [{"train_frac": 0.9, "test_size": 2}],
            [{"train_size": 9, "test_frac": 0.1}],
            [{"train_size": 9}],
        ]
    )
    def test_split(self, params: Dict[str, Any]) -> None:
        sdp = SimpleDataPartition(**params)
        for raw_data, truth in zip(ALL_DATASETS, RAW_SDP_RES):
            res = sdp.split(raw_data)
            self.assertTrue(res == truth)


class TestSimpleTimestampDataPartition(TestCase):
    # pyre-fixme
    @parameterized.expand(
        [
            (
                {"train_end": "2021-05-10", "test_start": "2021-05-13"},
                ALL_DATASETS[0],
                RAW_STDP_RES[0],
            ),
            (
                {"train_end": "2020-05-10", "test_start": "2020-05-13"},
                ALL_DATASETS[1],
                RAW_STDP_RES[1],
            ),
            (
                {"train_end": "2020-05-10", "test_start": "2021-05-13"},
                ALL_DATASETS[2],
                RAW_STDP_RES[2],
            ),
        ]
    )
    def test_split(self, params: Dict[str, Any], data: DataPartition, truth: List[TrainTestData]) -> None:
        stdp = SimpleTimestampDataPartition(**params)
        res = stdp.split(data)
        self.assertTrue(res == truth)


class TestRollingOriginDataParition(TestCase):

    # pyre-fixme
    @parameterized.expand(
        [
            (
                {
                    "start_train_frac": 0.5,
                    "test_frac": 0.2,
                    "expanding_steps": 2,
                },
                ALL_DATASETS[0],
                RAW_RODP_RES[0],
            ),
            (
                {
                    "start_train_frac": 0.5,
                    "test_frac": 0.5,
                    "expanding_steps": 1,
                },
                ALL_DATASETS[1],
                RAW_RODP_RES[1],
            ),
            (
                {
                    "start_train_frac": 0.3,
                    "test_frac": 0.05,
                    "expanding_steps": 3,
                    "constant_train_size": True,
                },
                ALL_DATASETS[2],
                RAW_RODP_RES[2],
            ),
            (
                {
                    "start_train_frac": 0.3,
                    "test_frac": 0.2,
                    "expanding_steps": 3,
                    "constant_train_size": True,
                    "window_frac": 0.1,
                },
                ALL_DATASETS[0],
                RAW_RODP_RES[3],
            ),
        ]
    )
    def test_split(self, params: Dict[str, Any], data: DataPartition, truth: List[TrainTestData]) -> None:
        rodp = RollingOriginDataParition(**params)
        res = rodp.split(data)
        self.assertTrue(res == truth)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from unittest import TestCase
from unittest.mock import MagicMock
from unittest.mock import patch

import matplotlib.pyplot as plt
import pandas as pd
from parameterized.parameterized import parameterized

from kats.consts import TimeSeriesData
from kats.data.utils import load_data
from kats.models.var import VARModel
from kats.models.var import VARParams

TEST_DATA = {
    "multivariate": {
        "ts": TimeSeriesData(load_data("multivariate_anomaly_simulated_data.csv")),
    },
    "multivariate_2": {
        "ts": TimeSeriesData(load_data("multi_ts.csv")),
    },
}


class testVARModel(TestCase):
    def test_fit_forecast(self, steps: int = 5) -> None:
        ts = TEST_DATA["multivariate_2"]["ts"]
        params = VARParams()
        train, truth = ts[:-steps], ts[-steps:]
        m = VARModel(train, params)
        m.fit()
        pred = m.predict(steps=steps)

        # check whether the time indices of each forecasted feature are the same
        index = [v.to_dataframe().time for _, v in pred.items()]
        self.assertTrue(all(x.equals(index[0]) for x in index))

        # check whether the values are close and shapes are correct
        truth = truth.to_dataframe().iloc[:, 1:]
        pred_forecast = pd.concat([v["fcst"].to_dataframe().iloc[:, 1:2] for _, v in pred.items()], axis=1)
        pred_forecast.columns = truth.columns
        self.assertTrue(truth.subtract(pred_forecast).values.max() < 5)

    def test_invalid_params(self) -> None:
        params = VARParams()
        input_data = TimeSeriesData(pd.DataFrame())
        with self.assertRaises(ValueError):
            m = VARModel(input_data, params)
            m.fit()
            m.predict(steps=30, include_history=True)

    # pyre-fixme[56]
    @parameterized.expand([[TEST_DATA["multivariate"]["ts"]]])
    @patch("pandas.concat")
    def test_predict_exception(self, ts: TimeSeriesData, mock_obj: MagicMock) -> None:
        mock_obj.side_effect = Exception
        with self.assertRaisesRegex(Exception, "^Failed to generate in-sample forecasts for historical data"):
            params = VARParams()
            m = VARModel(ts, params)
            m.fit()
            m.predict(steps=30, include_history=True)

    # pyre-fixme[56]
    @parameterized.expand(
        [
            [TEST_DATA["multivariate"]["ts"]],
            [TEST_DATA["multivariate_2"]["ts"]],
        ]
    )
    def test_predict_unfit(self, ts: TimeSeriesData) -> None:
        with self.assertRaises(ValueError):
            m = VARModel(ts, VARParams())
            m.predict(steps=30)

    # pyre-fixme[56]
    @parameterized.expand(
        [
            [TEST_DATA["multivariate"]["ts"]],
            [TEST_DATA["multivariate_2"]["ts"]],
        ]
    )
    def test_search_space(self, ts: TimeSeriesData) -> None:
        params = VARParams()
        params.validate_params()
        with self.assertRaises(NotImplementedError):
            VARModel.get_parameter_search_space()

    # pyre-fixme[56]
    @parameterized.expand([[TEST_DATA["multivariate"]["ts"]]])
    def test_str(self, ts: TimeSeriesData) -> None:
        result = str(VARModel(ts, VARParams()))
        self.assertEqual("VAR", result)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import logging
from datetime import datetime
from datetime import timedelta
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

try:
    import pymannkendall as mk

    _no_mk = False
except ImportError:
    _no_mk = True
from statsmodels.tsa.api import SimpleExpSmoothing

from kats.consts import TimeSeriesChangePoint
from kats.consts import TimeSeriesData
from kats.detectors.detector import Detector

"""Mann-Kendall (MK) Trend Detector Module

This module includes detectors based on the Mann-Kendall Test, which is a
non-parametric test for monotonic trends.  Right now, this module includes
detectors based on the original MK Test and the Multivariate MK Test.  Neither
detector has any distribution requirement for the data, but there should not be
any serial correlation.
"""

SEASON_FREQ_MAP: Dict[str, int] = {
    "daily": 1,
    "weekly": 7,
    "monthly": 30,
    "yearly": 365,
}


class MKChangePoint(TimeSeriesChangePoint):
    """Changepoint for MKDetector

    Attributes:

        start_time: Start time of the change.
        end_time: End time of the change.
        confidence: The confidence of the change point.
        detector_type: Detector, Type of detector changepoint is for.  Right
            now, this is always MKDetector.
        is_multivariate: boolean, Whether this is a changepoint for a
            multivariate time series.
        trend_direction: string, Direction of trend, either 'increasing' or
            'decreasing'.
        Tau: float or Dict, Kendall's Tau value for changepoint.
            This is a float in the univariate case and a Dict in the
                multivariate case.
    """

    def __init__(
        self,
        start_time: pd.Timestamp,
        end_time: pd.Timestamp,
        confidence: float,
        is_multivariate: bool,
        trend_direction: str,
        Tau: Union[float, Dict[str, Any]],
    ) -> None:
        super().__init__(start_time, end_time, confidence)
        if _no_mk:
            raise RuntimeError("requires pymannkendall to be installed")
        self._detector_type = MKDetector
        self._is_multivariate = is_multivariate
        self._trend_direction = trend_direction
        self._Tau = Tau

    @property
    def detector_type(self) -> Type[MKDetector]:
        return self._detector_type

    @property
    def is_multivariate(self) -> bool:
        return self._is_multivariate

    @property
    def trend_direction(self) -> str:
        return self._trend_direction

    @property
    def Tau(self) -> Union[float, Dict[str, Any]]:
        return self._Tau  # Tau is a dict in multivariate case

    def __repr__(self) -> str:
        return (
            f"MKChangePoint(start_time: {self._start_time}, end_time: "
            f"{self._end_time}, confidence: {self._confidence}, "
            f"detector_type: {self._detector_type}, is_multivariate: "
            f"{self._is_multivariate}, trend_direction: {self._trend_direction} "
            f"Tau: {self._Tau})"
        )


class MKDetector(Detector):
    """
    MKDetector (MK stands for Mann-Kendall) is a non-parametric statistical test
    used to determine whether there is a monotonic trend in a given time series.
    See https://vsp.pnnl.gov/help/vsample/Design_Trend_Mann_Kendall.htm for
    details.

    The basic idea is to check whether there is a monotonic trend based on a
    look back number of time steps (`window_size`).

    Parameters:

        data: `TimeSeriesData`, this is time series data at one-day granularity.
            This time series can be either univariate or multivariate.
            We require more than training_days points in each time series.
        threshold: `float`, threshold for trend intensity; higher threshold gives
            trend with high intensity (0.8 by default).  If we only want to use
            the p-value to determine changepoints, set threshold = 0.
        alpha: `float`, significance level (0.05 by default)
        multivariate: `bool`, whether the input time series is multivariate

    Example:
    --------
    >>> import pandas as pd
    >>> from kats.consts import TimeSeriesData
    >>> from kats.detectors.trend_mk import MKDetector
    >>> # read data and rename the two columns required by TimeSeriesData
    >>> # structure
    >>> data = pd.read_csv("../filename.csv") # demo file does not exist
    >>> TSdata = TimeSeriesData(data)
    >>> # create MKDetector with given data and params
    >>> d = MKDetector(data=TSdata)
    >>> # call detector method to fit model
    >>> detected_time_points = d.detector(window_size=20, direction="up")
    >>> # plot the results
    >>> d.plot(detected_time_points)
    """

    window_size: Optional[int] = None
    training_days: Optional[int] = None
    direction: Optional[str] = None
    freq: Optional[str] = None
    ts: Optional[pd.DataFrame] = None
    MK_statistics: Optional[pd.DataFrame] = None

    def __init__(
        self,
        data: Optional[TimeSeriesData] = None,
        threshold: float = 0.8,
        alpha: float = 0.05,
        multivariate: bool = False,
    ) -> None:
        # pyre-fixme[6]: Expected `TimeSeriesData` for 1st param but got
        #  `Optional[TimeSeriesData]`.
        super(MKDetector, self).__init__(data=data)
        if _no_mk:
            raise RuntimeError("requires pymannkendall to be installed")

        self.threshold = threshold
        self.alpha = alpha
        self.multivariate = multivariate
        self.__subtype__ = "trend_detector"

        # Assume univariate but multivariate data is detected
        if self.data is not None:
            if not self.data.is_univariate() and not self.multivariate:
                logging.warning("Using multivariate MK test for univariate data.")
                self.multivariate = True
            # Assume multivariate but univariate data is detected
            elif self.data.is_univariate() and self.multivariate:
                logging.warning("Using univariate MK test on multivariate data.")

    def _remove_seasonality(self, ts: pd.DataFrame, freq: Optional[str] = None) -> pd.DataFrame:
        """Remove seasonality in the time series using moving average."""

        if freq is None:
            return ts  # no seasonality
        else:
            ts = ts.rolling(window=SEASON_FREQ_MAP[freq]).mean()
        return ts

    def _smoothing(self, ts: pd.DataFrame) -> pd.DataFrame:
        """Remove noise in the time series using Holt-Winters model."""

        smoothed_ts = pd.DataFrame()
        for c in ts.columns:
            ts_c = ts[c].dropna()
            assert ts_c is not None
            with np.errstate(divide="raise"):
                try:
                    model = SimpleExpSmoothing(ts_c)
                    _fit = model.fit(smoothing_level=0.2, optimized=False)
                    smoothed_ts_tmp = _fit.predict(
                        start=ts_c.index[0],
                        end=ts_c.index[-1],
                    )
                    smoothed_ts = pd.concat([smoothed_ts, smoothed_ts_tmp.rename(c)], axis=1, copy=False)
                except FloatingPointError:
                    smoothed_ts = pd.concat([smoothed_ts, ts_c], axis=1, copy=False)
                    logging.debug("Your data does not have noise. No need for smoothing")

        return smoothed_ts

    def _preprocessing(self, ts: pd.DataFrame) -> Tuple[np.ndarray, int]:
        """Check and convert the dataframe ts to an numpy array.

        Args:
            ts: a time series dataframe with time as index.

        Returns:
            A numpy array of length n and the # of metrics.
        """
        window_size = self.window_size
        assert window_size is not None

        # takes only window_size days
        x = np.asarray(ts[-window_size:])
        dim = x.ndim

        # checks the dimension of the data
        if dim == 2:  # dim should always be 2
            (n, c) = x.shape  # n is # of obs = window_size; c is # of metrics
            if c == 1:  # univariate case
                dim = 1  # convert x from 2-dim array (n, 1) to 1-dim array (n,)
                x = x.flatten()
        else:
            msg = f"dim = 2 is expected but your data has dim = {dim}."
            raise ValueError(msg)

        return x, c

    def _drop_missing_values(self, x: np.ndarray) -> Tuple[np.ndarray, int]:
        """Drop the missing values in x."""

        if x.ndim == 1:  # univariate case with 1-dim array/ shape(n,)
            x = x[~(np.isnan(x))]
        else:  # multivariate case with 2-dim arrat/ shape (n, c)
            x = x[~np.isnan(x).any(axis=1)]
        return x, len(x)

    def _apply_threshold(self, trend: str, Tau: float) -> str:
        if abs(Tau) <= self.threshold:
            return "no trend"
        return trend

    def MKtest(self, ts: pd.DataFrame) -> Tuple[datetime, str, float, float]:
        """Performs the Mann-Kendall (MK) test for trend detection.

        (Mann 1945, Kendall 1975, Gilbert 1987)

        Args:
            ts: the dataframe of input data with time as index.
                This time series should not present seasonality for MK test.

        Returns:
            (tuple): tuple containing:

                anchor_date(datetime): the last time point in ts; the date for
                    which alert is triggered
                trend(str): tells the trend (decreasing, increasing, or no trend)
                p(float): p-value of the significance test
                Tau(float): Kendall Tau-b statistic (https://en.wikipedia.org/wiki/Kendall_rank_correlation_coefficient#Tau-b)
        """

        x, _ = self._preprocessing(ts)
        x, _ = self._drop_missing_values(x)

        anchor_date = ts.index[-1]

        mk_result = mk.original_test(x)
        trend, p, Tau = mk_result.trend, mk_result.p, mk_result.Tau
        trend = self._apply_threshold(trend, Tau)

        return anchor_date, trend, p, Tau

    def multivariate_MKtest(self, ts: pd.DataFrame) -> Tuple[datetime, str, float, Dict[str, Any]]:
        """Performs the Multivariate Mann-Kendall (MK) test.

        Proposed by R. M. Hirsch and J. R. Slack (1984).

        Args:
            ts: the dataframe of input data with time as index.  This time
                series should not present seasonality for MK test.

        Returns:
            (tuple): tuple containing:

                anchor_date(datetime): the last time point in ts; the date for
                    which alert is triggered.
                trend:_dict: tells the trend (decreasing, increasing, or no trend)
                    for each metric.
                p: p-value of the significance test.
                Tau_dict: Dictionary of Kendall Tau-b statistics for each univariate
                    time series, and Tau_dict["overall"] gives the Tau-b statistic
                    for the multivariate time series.
        """

        anchor_date = ts.index[-1]

        Tau_dict = {}  # contains score for individual cluster and overall
        trend_dict = {}  # contains trend for individual cluster and overall

        x, c = self._preprocessing(ts)

        multi_mk_result = mk.multivariate_test(x)
        trend, p, Tau = multi_mk_result.trend, multi_mk_result.p, multi_mk_result.Tau
        trend = self._apply_threshold(trend, Tau)

        Tau_dict["overall"] = Tau
        trend_dict["overall"] = trend

        for i in range(c):
            x_i, n = self._drop_missing_values(x[:, i])
            # individual Tau score and trend
            try:
                mk_result = mk.original_test(x_i)
                trend_i, Tau_i = mk_result.trend, mk_result.Tau
                trend_i = self._apply_threshold(trend_i, Tau_i)
                Tau_dict[ts.columns[i]] = Tau_i
                trend_dict[ts.columns[i]] = trend_i
            except ZeroDivisionError:
                Tau_dict[ts.columns[i]] = None
                trend_dict[ts.columns[i]] = None

        return anchor_date, trend_dict, p, Tau_dict

    def runDetector(self, ts: pd.DataFrame) -> Dict[str, Any]:
        """Runs MK test for a time point in the input data.

        Args:
            ts: the dataframe of input data with noise and seasonality removed.
                Its index is time.

        Returns:
            A dictionary consisting of MK test statistics for the anchor time
                point, including trend, p-value and Kendall Tau.
        """

        # run MK test
        if self.multivariate:
            anchor_date, trend, p, Tau = self.multivariate_MKtest(ts)
        else:
            anchor_date, trend, p, Tau = self.MKtest(ts)

        return {"ds": anchor_date, "trend_direction": trend, "p": p, "Tau": Tau}

    # pyre-fixme[14]: `detector` overrides method defined in `Detector` inconsistently.
    def detector(
        self,
        window_size: int = 20,
        training_days: Optional[int] = None,
        direction: str = "both",
        freq: Optional[str] = None,
    ) -> Sequence[MKChangePoint]:
        """Runs MK test sequentially.

        It finds the trend and calculates the related statistics for all time
        points in a given time series.

        Args:
            window_size: int, the number of look back days for checking trend
                persistence
            training_days: int, the number of days for time series smoothing;
                should be greater or equal to window_size
                If training_days is None, we will perform trend detection on the
                whole time series; otherwise, we will perform trend detection
                only for the anchor point using the previous training_days data.
            direction: string, the direction of the trend to be detected, choose
                from {"down", "up", "both"}
            freq: str, the type of seasonality shown in the time series,
                choose from {'daily', 'weekly','monthly','yearly'}
        """

        self.window_size = window_size
        self.training_days = training_days
        self.direction = direction
        self.freq = freq

        ts = self.data.to_dataframe().set_index("time")
        ts = ts.dropna(axis=1)
        ts.index = pd.DatetimeIndex(ts.index.values, freq=ts.index.inferred_freq)
        self.ts = ts

        if training_days is None:
            logging.info("Performing trend detection on the whole time series...")
            # check validity of the input value
            if len(ts) < window_size:
                raise ValueError(
                    f"For the whole time series analysis, data must have at "
                    f"least window_size={window_size} points."
                )

        else:
            logging.info(
                f"Performing trend detection for the anchor date {ts.index[-1]}"
                f" with training_days={training_days}..."
            )

            # check validity of the input value
            if training_days < window_size:
                raise ValueError(
                    f"For the anchor date analysis, training days should have "
                    f"at least window_size={window_size} points."
                )

            if len(ts) < training_days:
                raise ValueError(
                    f"For the anchor date analysis, data must have " f"at least training_days={training_days} points."
                )

        # save the trend detection results to dataframe MK_statistics
        MK_statistics = pd.DataFrame(columns=["ds", "trend_direction", "p", "Tau"])

        if training_days is not None:  # anchor date analysis for real-time setting
            # only look back training_days for noise and seasonality removal
            start = ts.index[-1] - timedelta(days=training_days)
            ts = ts.loc[start : ts.index[-1], :]
            # deseasonalization
            ts_deseas = self._remove_seasonality(ts, freq=self.freq)
            ts_smoothed = self._smoothing(ts_deseas)  # smoothing
            # append MK statistics to MK_statistics dataframe
            MK_statistics = MK_statistics.append(
                # pyre-ignore[6]: Expected `Union[Dict[Union[int, str], typing.Any], L...
                self.runDetector(ts=ts_smoothed),
                ignore_index=True,
            )

        else:
            # use the whole time series for for noise and seasonality removal
            ts_deseas = self._remove_seasonality(ts, freq=freq)
            ts_smoothed = self._smoothing(ts_deseas)

            # run detector sequentially with sliding_window for the whole time
            # series
            for t in ts_smoothed.index[window_size:]:
                # look back window_size day for trend detection
                ts_tmp = ts_smoothed.loc[:t, :]
                # append MK statistics to MK_statistics dataframe
                MK_statistics = MK_statistics.append(
                    # pyre-ignore[6]: Expected `Union[Dict[Union[int, str], typing.Any...
                    self.runDetector(ts=ts_tmp),
                    ignore_index=True,
                )

        self.MK_statistics = MK_statistics

        # take the subset for detection with specified trend_direction
        MK_results = self.get_MK_results(MK_statistics=MK_statistics, direction=direction)

        return self._convert_detected_tps(MK_results)

    def get_MK_results(self, MK_statistics: pd.DataFrame, direction: str) -> pd.DataFrame:
        """Obtain a subset of MK_statistics given the desired direction"""

        if direction not in ["up", "down", "both"]:
            raise ValueError("direction should be chosen from {'up', 'down', 'both'}")

        if self.multivariate:
            trend_df = pd.DataFrame.from_dict(list(MK_statistics.trend_direction))
            overall_trend = trend_df["overall"]

            if direction == "down":
                MK_results = MK_statistics.loc[overall_trend == "decreasing", :]
            elif direction == "up":
                MK_results = MK_statistics.loc[overall_trend == "increasing", :]
            elif direction == "both":
                MK_results = MK_statistics.loc[overall_trend != "no trend", :]
        else:
            if direction == "down":
                MK_results = MK_statistics.loc[MK_statistics["trend_direction"] == "decreasing", :]
            elif direction == "up":
                MK_results = MK_statistics.loc[MK_statistics["trend_direction"] == "increasing", :]
            elif direction == "both":
                MK_results = MK_statistics.loc[MK_statistics["trend_direction"] != "no trend", :]
        # pyre-fixme[61]: `MK_results` may not be initialized here.
        return MK_results

    def _convert_detected_tps(self, MK_results: pd.DataFrame) -> List[MKChangePoint]:
        """Convert the dataframe of detected_tps and Tau into desired format."""

        converted = []

        for _index, row in MK_results.iterrows():
            t = row["ds"]
            detected_time_point = MKChangePoint(
                start_time=t,
                end_time=t,
                confidence=1 - row["p"],
                is_multivariate=self.multivariate,
                trend_direction=row["trend_direction"],
                Tau=row["Tau"],
            )
            converted.append(detected_time_point)

        return converted

    def get_MK_statistics(self) -> pd.DataFrame:
        """Get the dataframe of MK_statistics."""
        MK_statistics = self.MK_statistics
        if MK_statistics is None:
            raise ValueError("Call detector() first.")
        return MK_statistics

    def get_top_k_metrics(self, time_point: datetime, top_k: Optional[int] = None) -> pd.DataFrame:
        """Get k metrics that show the most significant trend at a time point.

        Works only for multivariate data.

        Args:
            time_point: the time point to be investigated.
            top_k: the number of top metrics.

        Returns:
            a dataframe consists of top_k metrics and their corresponding
                Kendall Tau and trend.
        """

        Tau_df, trend_df = self._metrics_analysis()
        Tau_df = Tau_df.melt(id_vars=["ds"], var_name="metric", value_name="Tau")
        trend_df = trend_df.melt(id_vars=["ds"], var_name="metric", value_name="trend_direction")

        if self.training_days is not None:
            time_point = self.data.time.iloc[-1]
            # time_point default to the only anchor date for real-time detection

        # obtain the Tau for all metrics at the time point
        Tau_df_tp = Tau_df.loc[Tau_df["ds"] == time_point, :]
        trend_df_tp = trend_df.loc[trend_df["ds"] == time_point, :]
        MK_statistics_tp = pd.merge(Tau_df_tp, trend_df_tp, copy=False)

        # sort the metrics according to their Tau
        if self.direction == "down":
            top_metrics = MK_statistics_tp.reindex(MK_statistics_tp.Tau.sort_values(axis=0).index)
        elif self.direction == "up":
            top_metrics = MK_statistics_tp.reindex(MK_statistics_tp.Tau.sort_values(axis=0, ascending=False).index)
        else:
            assert self.direction == "both"
            top_metrics = MK_statistics_tp.reindex(
                MK_statistics_tp.Tau.abs().sort_values(axis=0, ascending=False).index
            )

        if top_k is None:
            # if top_k not specified, return all metrics ranked by Tau
            return top_metrics

        return top_metrics.iloc[:top_k]

    def _metrics_analysis(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        if not self.multivariate:
            raise ValueError("Your data is not multivariate.")
        MK_statistics = self.MK_statistics
        assert MK_statistics is not None

        # obtain the Tau for all metrics at all time points
        Tau_df = pd.DataFrame.from_dict(list(MK_statistics.Tau))
        Tau_df["ds"] = MK_statistics.ds
        Tau_df = Tau_df.drop(["overall"], axis=1)  # remove overall score

        trend_df = pd.DataFrame.from_dict(list(MK_statistics.trend_direction))
        trend_df["ds"] = MK_statistics.ds
        trend_df = trend_df.drop(["overall"], axis=1)  # remove overall trend

        return Tau_df, trend_df

    def plot(
        self,
        detected_time_points: Sequence[MKChangePoint],
        ax: Optional[plt.Axes] = None,
        figsize: Optional[Tuple[int, int]] = None,
        changepoint_color: str = "red",
        **kwargs: Any,
    ) -> plt.Axes:
        """Plots the original time series data and the detected time points.

        Args:
            detected_time_points: the detected change points.
            ax: optional matplotlib Axes. If None, creates one.
            figsize: optional figure size. If None, uses (14, 5).
            changepoint_color: the color to use for changepoint vertical lines.
        """
        ts = self.ts
        if ts is None:
            raise ValueError("detector() must be called before plot()")

        with pd.option_context("plotting.matplotlib.register_converters", True):
            if ax is None:
                if figsize is None:
                    figsize = (14, 5)
                _, ax = plt.subplots(figsize=figsize)

            ax.plot(ts.index, ts.values)

            trend_annotated = False
            for t in detected_time_points:
                ax.axvline(x=t.start_time, color=changepoint_color)
                trend_annotated = True
            if not trend_annotated:
                logging.warning("No trend detected!")
            return ax



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
import pandas as pd
from statsmodels.tsa.statespace.mlemodel import MLEResults
from statsmodels.tsa.statespace.sarimax import SARIMAX

from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.models.model import Model
from kats.utils.parameter_tuning_utils import get_default_sarima_parameter_search_space

ArrayLike = np.ndarray


class SARIMAParams(Params):
    """Parameter class for SARIMA model

    This is the parameter class for SARIMA model, it contains all necessary
    parameters as defined in SARIMA model implementation:
    https://www.statsmodels.org/stable/generated/statsmodels.tsa.statespace.sarimax.SARIMAX.html.

    Attributes:
        p: An integer for trend autoregressive (AR) order.
        d: An integer for trend difference order.
        q: An integer for trend moving average (MA) order.
        exog: Optional; An array of exogenous regressors.
        seasonal_order: Optional; A tuple for (P,D,Q,s) order of the seasonal component
            for AR order, difference order, MA order, and periodicity. Default is
            (0,0,0,0).
        trend: Optional; A string or an iterable for deterministic trend. Can be 'c'
            (constant), 't' (linear trend with time), 'ct' (both constant and linear
            trend), or an iterable of integers defining the non-zero polynomial
            exponents to include. Default is None (not to include trend).
        measurement_error: Optional; A boolean to specify whether or not to assume the
            observed time series were measured with error. Default is False.
        time_varying_regression: Optional; A boolean to specify whether or not
            coefficients on the regressors (if provided) are allowed to vary over time.
            Default is False.
        mle_regression: Optional; A boolean to specify whether or not to estimate
            coefficients of regressors as part of maximum likelihood estimation or
            through Kalman filter. If time_varying_regression is True, this must be
            set to False. Default is True.
        simple_differencing: Optional; A boolean to specify whether or not to use
            partially conditional maximum likelihood estimation. See
            https://www.statsmodels.org/stable/generated/statsmodels.tsa.statespace.sarimax.SARIMAX.html
            for more details. Default is False.
        enforce_stationarity: Optional; A boolean to specify whether or not to
            transform the AR parameters to enforce stationarity in AR component.
            Default is True.
        enforce_invertibility: Optional; A boolean to specify whether or not to
            transform the MA parameters to enforce invertibility in MA component.
            Default is True.
        hamilton_representation: Optional; A boolean to specify whether or not to
            use the Hamilton representation or the Harvey representation (if False).
            Default is False.
        concentrate_scale: Optional; A boolean to specify whether or not to concentrate
            the scale (variance of the error term) out of the likelihood. Default
            is False.
        trend_offset: Optional; An integer for the offset at which to start time
            trend value. Default is 1.
        use_exact_diffuse: Optional; A boolean to specify whether or not to use exact
            diffuse initialization for non-stationary states. Default is False.
    """

    def __init__(
        self,
        p: int,
        d: int,
        q: int,
        exog: Optional[ArrayLike] = None,
        seasonal_order: Tuple[int, ...] = (0, 0, 0, 0),
        trend: Optional[str] = None,
        measurement_error: bool = False,
        time_varying_regression: bool = False,
        mle_regression: bool = True,
        simple_differencing: bool = False,
        enforce_stationarity: bool = True,
        enforce_invertibility: bool = True,
        hamilton_representation: bool = False,
        concentrate_scale: bool = False,
        trend_offset: int = 1,
        use_exact_diffuse: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__()
        self.p = p
        self.d = d
        self.q = q
        self.exog = exog
        self.seasonal_order = seasonal_order
        self.trend = trend
        self.measurement_error = measurement_error
        self.time_varying_regression = time_varying_regression
        self.mle_regression = mle_regression
        self.simple_differencing = simple_differencing
        self.enforce_stationarity = enforce_stationarity
        self.enforce_invertibility = enforce_invertibility
        self.hamilton_representation = hamilton_representation
        self.concentrate_scale = concentrate_scale
        self.trend_offset = trend_offset
        self.use_exact_diffuse = use_exact_diffuse
        logging.debug(
            "Initialized SARIMAParams with parameters. " f"p:{p}, d:{d}, q:{q},seasonal_order:{seasonal_order}"
        )

    def validate_params(self) -> None:
        """Not implemented."""

        logging.info("Method validate_params() is not implemented.")
        pass


class SARIMAModel(Model[SARIMAParams]):
    """Model class for SARIMA.

    This class provides fit, predict and plot methods for SARIMA model.

    Attributes:
        data: :class:`kats.consts.TimeSeriesData` object for input time series.
        params: :class:`SARIMAParams` for model parameters.
    """

    start_params: Optional[np.ndarray] = None
    transformed: Optional[bool] = None
    includes_fixed: Optional[bool] = None
    cov_type: Optional[str] = None
    cov_kwds: Optional[Dict[str, Any]] = None
    method: Optional[str] = None
    maxiter: Optional[int] = None
    full_output: Optional[bool] = None
    disp: Optional[bool] = None
    callback: Optional[Callable[[np.ndarray], None]] = None
    return_params: Optional[bool] = None
    optim_score: Optional[str] = None
    optim_complex_step: Optional[bool] = None
    optim_hessian: Optional[str] = None
    low_memory: Optional[bool] = None
    model: Optional[MLEResults] = None
    include_history: bool = False
    alpha: float = 0.05
    fcst_df: Optional[pd.DataFrame] = None
    freq: Optional[float] = None
    y_fcst: Optional[np.ndarray] = None
    y_fcst_lower: Optional[np.ndarray] = None
    y_fcst_upper: Optional[np.ndarray] = None
    dates: Optional[pd.DatetimeIndex] = None

    def __init__(
        self,
        data: TimeSeriesData,
        params: SARIMAParams,
    ) -> None:
        super().__init__(data, params)
        # pyre-fixme[16]: `Optional` has no attribute `value`.
        if not isinstance(self.data.value, pd.Series):
            msg = f"Only support univariate time series, but get {self.data.value}."
            logging.error(msg)
            raise ValueError(msg)

    def fit(
        self,
        start_params: Optional[np.ndarray] = None,
        transformed: bool = True,
        includes_fixed: bool = False,
        cov_type: Optional[str] = None,
        cov_kwds: Optional[Dict[str, Any]] = None,
        method: str = "lbfgs",
        maxiter: int = 50,
        full_output: bool = True,
        disp: bool = False,
        callback: Optional[Callable[[np.ndarray], None]] = None,
        return_params: bool = False,
        optim_score: Optional[str] = None,
        optim_complex_step: bool = True,
        optim_hessian: Optional[str] = None,
        low_memory: bool = False,
    ) -> None:
        """Fit SARIMA model by maximum likelihood via Kalman filter.

        See reference
        https://www.statsmodels.org/stable/generated/statsmodels.tsa.statespace.sarimax.SARIMAX.fit.html#statsmodels.tsa.statespace.sarimax.SARIMAX.fit
        for more details.

        Args:
            start_params: Optional; An array_like object for the initial guess of the
                solution for the loglikelihood maximization.
            transformed: Optional; A boolean to specify whether or not start_params
                is already transformed. Default is True.
            includes_fixed: Optional; A boolean to specify whether or not start_params
                includes the fixed parameters in addition to the free parameters.
                Default is False.
            cov_type: Optional; A string for the method for calculating the covariance
                matrix of parameter estimates. Can be 'opg' (outer product of gradient
                estimator), 'oim' (observed information matrix estimator), 'approx'
                (observed information matrix estimator), 'robust' (approximate
                (quasi-maximum likelihood) covariance matrix), or 'robust_approx'.
                Default is 'opg' when memory conservation is not used, otherwise
                default is ‘approx’.
            cov_kwds: Optional; A dictionary of arguments for covariance matrix
                computation. See reference for more details.
            method: Optional; A string for solver from scipy.optimize to be used.
                Can be 'newton', 'nm', 'bfgs', 'lbfgs', 'powell', 'cg', 'ncg' or
                'basinhopping'. Default is 'lbfgs'.
            maxiter: Optional; An integer for the maximum number of iterations to
                perform. Default is 50.
            full_output: Optional; A boolean to specify whether or not to have all
                available output in the Results object’s mle_retvals attribute.
                Default is True.
            disp: Optional; A boolean to specify whether or not to print convergence
                messages. Default is False.
            callback: Optional; A callable object to be called after each iteration.
                Default is None.
            return_params: Optional; A boolean to specify whether or not to return
                only the array of maximizing parameters. Default is False.
            optim_score: Optional; A string for the method by which the score vector
                is calculated. Can be 'harvey', 'approx' or None. Default is None.
            optim_complex_step: Optional; A boolean to specify whether or not to use
                complex step differentiation when approximating the score. Default
                is True.
            optim_hessian: Optional; A string for the method by which the Hessian is
                numerically approximated. Can be 'opg', 'oim', 'approx'. Default is
                None.
            low_memory: Optional; A boolean to specify whether or not to reduce memory
                usage. If True, some features of the results object will not be
                available. Default is False.

        Returns:
            None.
        """

        logging.debug("Call fit() method")
        self.start_params = start_params
        self.transformed = transformed
        self.includes_fixed = includes_fixed
        self.cov_type = cov_type
        self.cov_kwds = cov_kwds
        self.method = method
        self.maxiter = maxiter
        self.full_output = full_output
        self.disp = disp
        self.callback = callback
        self.return_params = return_params
        self.optim_score = optim_score
        self.optim_complex_step = optim_complex_step
        self.optim_hessian = optim_hessian
        self.low_memory = low_memory

        logging.info("Created SARIMA model.")
        sarima = SARIMAX(
            # pyre-fixme[16]: `Optional` has no attribute `value`.
            self.data.value,
            order=(self.params.p, self.params.d, self.params.q),
            exog=self.params.exog,
            seasonal_order=self.params.seasonal_order,
            trend=self.params.trend,
            measurement_error=self.params.measurement_error,
            time_varying_regression=self.params.time_varying_regression,
            mle_regression=self.params.mle_regression,
            simple_differencing=self.params.simple_differencing,
            enforce_stationarity=self.params.enforce_stationarity,
            enforce_invertibility=self.params.enforce_invertibility,
            hamilton_representation=self.params.hamilton_representation,
            concentrate_scale=self.params.concentrate_scale,
            trend_offset=self.params.trend_offset,
            use_exact_diffuse=self.params.use_exact_diffuse,
        )
        self.model = sarima.fit(
            start_params=self.start_params,
            transformed=self.transformed,
            includes_fixed=self.includes_fixed,
            cov_type=self.cov_type,
            cov_kwds=self.cov_kwds,
            method=self.method,
            maxiter=self.maxiter,
            full_output=self.full_output,
            disp=self.disp,
            callback=self.callback,
            return_params=self.return_params,
            optim_score=self.optim_score,
            optim_complex_step=self.optim_complex_step,
            optim_hessian=self.optim_hessian,
            low_memory=self.low_memory,
        )
        logging.info("Fitted SARIMA.")

    # pyre-fixme[14]: `predict` overrides method defined in `Model` inconsistently.
    # pyre-fixme[15]: `predict` overrides method defined in `Model` inconsistently.
    def predict(
        self,
        steps: int,
        exog: Optional[ArrayLike] = None,
        include_history: bool = False,
        alpha: float = 0.05,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """Predict with fitted SARIMA model.

        Args:
            steps: An integer for forecast steps.
            include_history: Optional; A boolearn to specify whether to include
                historical data. Default is False.
            alpha: A float for confidence level. Default is 0.05.
            exog: A numpy array of exogenous values to be passed to forecast

        Returns:
            A :class:`pandas.DataFrame` of forecasts and confidence intervals.
        """
        model = self.model
        if model is None:
            msg = "Call fit() before predict()."
            logging.error(msg)
            raise ValueError(msg)
        if (self.params.exog is not None) and (exog is None):
            msg = (
                "SARIMA model was initialized with exogenous variables. Exogenous "
                "variables must be used to predict. use `exog=`"
            )
            logging.error(msg)
            raise ValueError(msg)

        logging.debug(f"Call predict() with parameters. steps:{steps}, kwargs:{kwargs}")
        self.include_history = include_history
        # pyre-fixme[16]: `Optional` has no attribute `infer_freq_robust`.
        self.freq = kwargs.get("freq", self.data.infer_freq_robust())
        self.alpha = alpha

        fcst = model.get_forecast(steps, exog=exog)

        logging.info("Generated forecast data from SARIMA model.")
        logging.debug(f"Forecast data: {fcst}")

        if fcst.predicted_mean.isna().sum() == steps:
            msg = "SARIMA model fails to generate forecasts, i.e., all forecasts are " "NaNs."
            logging.error(msg)
            raise ValueError(msg)

        self.y_fcst = fcst.predicted_mean
        pred_interval = fcst.conf_int(alpha=self.alpha)

        if pred_interval.iloc[0, 0] < pred_interval.iloc[0, 1]:
            self.y_fcst_lower = np.array(pred_interval.iloc[:, 0])
            self.y_fcst_upper = np.array(pred_interval.iloc[:, 1])
        else:
            self.y_fcst_lower = np.array(pred_interval.iloc[:, 1])
            self.y_fcst_upper = np.array(pred_interval.iloc[:, 0])

        # pyre-fixme[16]: `Optional` has no attribute `time`.
        last_date = self.data.time.max()
        dates = pd.date_range(start=last_date, periods=steps + 1, freq=self.freq)

        self.dates = dates[dates != last_date]  # Return correct number of periods

        if include_history:
            # generate historical fit
            history_fcst = model.get_prediction(0)
            history_ci = history_fcst.conf_int()
            if ("lower" in history_ci.columns[0]) and ("upper" in history_ci.columns[1]):
                ci_lower_name, ci_upper_name = (
                    history_ci.columns[0],
                    history_ci.columns[1],
                )
            else:
                msg = "Error when getting prediction interval from statsmodels SARIMA API"
                logging.error(msg)
                raise ValueError(msg)
            self.fcst_df = fcst_df = pd.DataFrame(
                {
                    "time": np.concatenate((pd.to_datetime(self.data.time), self.dates)),
                    "fcst": np.concatenate((history_fcst.predicted_mean, self.y_fcst)),
                    "fcst_lower": np.concatenate((history_ci[ci_lower_name], self.y_fcst_lower)),
                    "fcst_upper": np.concatenate((history_ci[ci_upper_name], self.y_fcst_upper)),
                },
                copy=False,
            )

            # the first k elements of the fcst and lower/upper are not legitmate
            # thus we need to assign np.nan to avoid confusion
            # k = max(p, d, q) + max(P, D, Q) * seasonal_order + 1
            k = (
                max(self.params.p, self.params.d, self.params.q)
                + max(self.params.seasonal_order[0:3]) * self.params.seasonal_order[3]
                + 1
            )

            fcst_df.loc[0:k, ["fcst", "fcst_lower", "fcst_upper"]] = np.nan
        else:
            self.fcst_df = fcst_df = pd.DataFrame(
                {
                    "time": self.dates,
                    "fcst": self.y_fcst,
                    "fcst_lower": self.y_fcst_lower,
                    "fcst_upper": self.y_fcst_upper,
                },
                copy=False,
            )

        logging.debug(f"Return forecast data: {fcst_df}")
        return fcst_df

    def __str__(self) -> str:
        return "SARIMA"

    @staticmethod
    # pyre-fixme[15]: `get_parameter_search_space` overrides method defined in
    #  `Model` inconsistently.
    def get_parameter_search_space() -> List[Dict[str, Any]]:
        """Get default SARIMA parameter search space.

        Returns:
            A dictionary representing the default SARIMA parameter search space.
        """
        return get_default_sarima_parameter_search_space()
